# -*- coding: utf-8 -*-
#
# This file is part of the qpageview package.
#
# Copyright (c) 2010 - 2019 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.


"""
Manages and positions a group of Page instances.
"""


import copy
import itertools
import math

from PyQt5.QtCore import QMargins, QPoint, QPointF, QRect, QSize, Qt

from . import rectangles
from . import util
from .constants import (
    FixedScale,
    FitWidth,
    FitHeight,
    FitBoth,

    Rotate_0,
    Rotate_90,
    Rotate_180,
    Rotate_270,

    Horizontal,
    Vertical,
)


class PageRects(rectangles.Rectangles):
    def get_coords(self, page):
        return page.geometry().getCoords()


class PageLayout(util.Rectangular, list):
    """Manages page.Page instances with a list-like api.

    You can iterate over the layout itself, which yields all Page instances.

    The following instance attributes are used, with these class-level
    defaults::

        zoomFactor = 1.0
        dpiX = 72.0
        dpiY = 72.0
        rotation = Rotate_0
        orientation = Vertical
        alignment = Qt.AlignCenter

    The layout has margins around each page, accessible via pageMargins(), and
    margins around the whole layout, accessible via margins(). Both have class
    level defaults as a tuple, but they are converted to a QMargins object for
    the layout instance when first accessed via the margins() and pageMargins()
    methods::

        _margins = (6, 6, 6, 6)
        _pageMargins = (0, 0, 0, 0)

        spacing = 8             # pixels between pages

        x = 0                   # x, y, width and height are set by update()
        y = 0
        width = 0
        height = 0

        continuousMode = True   # whether to show all pages

    The actual layout is done by a LayoutEngine in the engine attribute.
    After having changed pages, engine or layout attributes, call update() to
    update the layout.

    """

    _margins = (6, 6, 6, 6)
    _pageMargins = (0, 0, 0, 0)
    spacing = 8
    zoomFactor = 1.0
    dpiX = 72.0
    dpiY = 72.0
    rotation = Rotate_0
    orientation = Vertical
    alignment = Qt.AlignCenter

    continuousMode = True
    currentPageSet = 0  # used in non-continuous mode

    _rects = None

    def __bool__(self):
        """Always return True."""
        return True

    def count(self):
        """Return the number of Page instances."""
        return len(self)

    def empty(self):
        """Return True if there are zero pages."""
        return len(self) == 0

    def setMargins(self, margins):
        """Sets our margins to a QMargins object."""
        self._m = margins

    def margins(self):
        """Return our margins as a QMargins object, intialized from _margins"""
        try:
            return self._m
        except AttributeError:
            self._m = QMargins(*self._margins)
            return self._m

    def setPageMargins(self, margins):
        """Sets our page margins to a QMargins object."""
        self._pm = margins

    def pageMargins(self):
        """Return our page margins as a QMargins object, intialized from _pageMargins"""
        try:
            return self._pm
        except AttributeError:
            self._pm = QMargins(*self._pageMargins)
            return self._pm

    def _pageRects(self):
        """(Internal) Return the PageRects object for quickly finding pages."""
        if self._rects:
            return self._rects
        r = self._rects = PageRects(self.displayPages())
        return r

    def pageAt(self, point):
        """Return the page that contains the given QPoint.

        If the point is not on any page, None is returned.

        """
        for page in self._pageRects().at(point.x(), point.y()):
            return page

    def pagesAt(self, rect):
        """Yield the pages touched by the given QRect.

        The pages are in undefined order.

        """
        for page in self._pageRects().intersecting(*rect.getCoords()):
            yield page

    def nearestPageAt(self, point):
        """Return the page at the shortest distance from the given point.

        The returned page does not contain the point. (Use pageAt() for that.)
        If there are no pages outside the point, None is returned.

        """
        return self._pageRects().nearest(point.x(), point.y())

    def defaultWidth(self, page):
        """Return the default width of the page."""
        if (page.rotation + self.rotation) & 1:
            return page.pageHeight * page.scaleY / page.dpi
        else:
            return page.pageWidth * page.scaleX / page.dpi

    def defaultHeight(self, page):
        """Return the default height of the page."""
        if (page.rotation + self.rotation) & 1:
            return page.pageWidth * page.scaleX / page.dpi
        else:
            return page.pageHeight * page.scaleY / page.dpi

    def widestPage(self):
        """Return the page with the largest default width, if any."""
        if self.count():
            return max(self, key=self.defaultWidth)

    def highestPage(self):
        """Return the page with the largest default height, if any."""
        if self.count():
            return max(self, key=self.defaultHeight)

    def fit(self, size, mode):
        """Fits the layout in the given size (QSize) and ViewMode."""
        self.engine.fit(self, size, mode)

    def zoomsToFit(self):
        """Return True if the layout engine changes the zoomFactor to fit."""
        return self.engine.zoomToFit

    def update(self):
        """Compute the size of all pages and updates their positions.
        Finally set our own size.

        You should call this after having added or deleted pages or after
        having changed the scale, dpi, zoom factor, spacing or margins.

        This function returns True if the total geometry has changed.

        """
        self._rects = None
        self.updatePageSizes()
        if self.count():
            self.engine.updatePagePositions(self)
        geometry = self.computeGeometry()
        changed = self.geometry() != geometry
        self.setGeometry(geometry)
        return changed

    def updatePageSizes(self):
        """Compute the correct size of every Page."""
        for page in self:
            page.computedRotation = (page.rotation + self.rotation) & 3
            page.updateSize(self.dpiX, self.dpiY, self.zoomFactor)

    def computeGeometry(self):
        """Return the total geometry (position and size) of the layout.

        In most cases the implementation of this method is sufficient: it
        computes the bounding rectangle of all Pages and adds the margin.

        """
        r = QRect()
        for page in self.displayPages():
            r |= page.geometry()
        return r + self.margins() + self.pageMargins()

    def pos2offset(self, pos):
        """Return a three-tuple (index, x, y).

        The index refers to a page in the layout, or nowhere if -1. The x and y
        refer to a spot on the page (or layout if empty) in the range 0..1.
        You can use it to store a certain position and restore it after
        changing the zoom e.g.

        """
        page = self.pageAt(pos) or self.nearestPageAt(pos)
        if page:
            pos = pos - page.pos()
            w = page.width
            h = page.height
            i = self.index(page)
        else:
            w = self.width
            h = self.height
            i = -1
        x = pos.x() / w
        y = pos.y() / h
        return (i, x, y)

    def offset2pos(self, offset):
        """Return the pos on the layout for the specified offset.

        The offset is a three-tuple like returned by pos2offset().

        """
        i, x, y = offset
        if i < 0 or i >= len(self):
            pos = QPoint(0, 0)
            w = self.width
            h = self.height
        else:
            page = self[i]
            pos = page.pos()
            w = page.width
            h = page.height
        return pos + QPoint(round(x * w), round(y * h))

    def displayPages(self):
        """Return the pages that are to be displayed."""
        return self[self.currentPageSetSlice()]

    def currentPageSetSlice(self):
        """Return a slice object describing the current page set."""
        if not self.continuousMode:
            num = self.currentPageSet
            count = self.pageSetCount()
            # make sure a valid slice is returned
            if num and num >= count:
                num = self.currentPageSet = count - 1
            p = 0
            s = 0
            for count, length in self.pageSets():
                if p + count <= num:
                    p += count
                    s += count * length
                    continue
                count = num - p
                s += count * length
                return slice(s, s + length)
        return slice(0, self.count())

    def pageSets(self):
        """Return a list of (count, length) tuples.

        Every count is the number of page sets of that length. The list is
        created by the LayoutEngine.pageSets() method.

        """
        return self.engine.pageSets(self.count())

    def pageSetCount(self):
        """Return the number of page sets."""
        return sum(count for count, length in self.pageSets())

    def pageSet(self, index):
        """Return the page set containing page at index."""
        s = 0   # the index at the start of the last page set
        p = 0   # the page set
        for count, length in self.pageSets():
            if s + count * length < index:
                s += count * length
                p += count
                continue
            return p + (index - s) // length
        return 0    # happens with empty layout


class LayoutEngine:
    """A LayoutEngine takes care of the actual layout process.

    A PageLayout has its LayoutEngine in the `engine` attribute.  Putting this
    functionality in a separate object makes it easier to alter the behaviour
    of a layout without changing all the user-set options and added Pages.

    The default implementation of LayoutEngine puts pages in a horizontal
    or vertical row.

    You can override grid() to implement a different behaviour, and you
    can override pageSets() to get a different behaviour in non-continuous mode.

    If there are multiple rows or columns, every row is as high as the
    highest page it contains, and every column is as wide as its widest page.
    You can set the attributes evenWidths and/or evenHeights to True if you
    want all columns to have the same width, and/or respectively, the rows the
    same height.

    """

    zoomToFit = True        # True means: engine changes the zoomFactor to fit
    orientation = None      # None means: use layout orientation

    evenWidths = False
    evenHeights = False

    def grid(self, layout):
        """Return a three-tuple (ncols, nrows, prepend).

        ncols is the number of columns the layout will contain, nrows the
        number of rows; and prepend if the number of empty positions that the
        layout wants, when the first row has less pages.

        """
        if layout.orientation == Vertical:
            return 1, layout.count(), 0
        else:
            return layout.count(), 1, 0

    def pages(self, layout, ncols, nrows, prepend=0):
        """Yield the layout's pages in a grid: (page, (x, y)).

        If prepend > 0, that number of first grid positions will remain unused.
        This can be used for layouts that have less pages in the first row.

        """
        if (self.orientation or layout.orientation) == Vertical:
            gen = ((col, row) for col in range(ncols) for row in range(nrows))
        else:
            gen = ((col, row) for row in range(nrows) for col in range(ncols))
        if prepend:
            for i in itertools.islice(gen, prepend):
                pass # skip unused positions
        return zip(layout, gen)

    def dimensions(self, layout, ncols, nrows, prepend=0):
        """Return two lists: columnwidths and rowheights.

        The width and height are page dimensions, without page margin.

        """
        colwidths = [0] * ncols
        rowheights = [0] * nrows
        for page, (col, row) in self.pages(layout, ncols, nrows, prepend):
            colwidths[col] = max(colwidths[col], page.width)
            rowheights[row] = max(rowheights[row], page.height)
        if self.evenWidths:
            colwidths = [max(colwidths)] * ncols
        if self.evenHeights:
            rowheights = [max(rowheights)] * nrows
        return colwidths, rowheights

    def updatePagePositions(self, layout):
        """Performs the positioning of the pages. Don't call on empty layout."""
        ncols, nrows, prepend = self.grid(layout)
        colwidths, rowheights = self.dimensions(layout, ncols, nrows, prepend)

        m = layout.margins()
        pm = layout.pageMargins()
        pmh = pm.left() + pm.right()        # horizontal page margin
        pmv = pm.top() + pm.bottom()        # vertical page margin

        # accumulate for column and row offsets, adding spacing
        xoff = [m.left() + pm.left()] + colwidths[:-1]
        yoff = [m.top() + pm.top()] + rowheights[:-1]
        for i in range(1, ncols):
            xoff[i] += xoff[i-1] + layout.spacing + pmh
        for i in range(1, nrows):
            yoff[i] += yoff[i-1] + layout.spacing + pmv
        # and go for positioning!
        for page, (col, row) in self.pages(layout, ncols, nrows, prepend):
            x, y = util.align(page.width, page.height, colwidths[col], rowheights[row], layout.alignment)
            page.x = xoff[col] + x
            page.y = yoff[row] + y

    def fit(self, layout, size, mode):
        """Called by PageLayout.fit()."""
        if mode and layout.count():
            zoomfactors = []
            if mode & FitWidth:
                zoomfactors.append(self.zoomFitWidth(layout, size.width()))
            if mode & FitHeight:
                zoomfactors.append(self.zoomFitHeight(layout, size.height()))
            layout.zoomFactor = min(zoomfactors)

    def zoomFitWidth(self, layout, width):
        """Return the zoom factor this layout would need to fit in the width.

        This method is called by fit(). The default implementation returns a
        suitable zoom factor for the widest Page.

        """
        m, p = layout.margins(), layout.pageMargins()
        width -= m.left() + m.right() + p.left() + p.right()
        return layout.widestPage().zoomForWidth(width, layout.rotation, layout.dpiX)

    def zoomFitHeight(self, layout, height):
        """Return the zoom factor this layout would need to fit in the height.

        This method is called by fit(). The default implementation returns a
        suitable zoom factor for the highest Page.

        """
        m, p = layout.margins(), layout.pageMargins()
        height -= m.top() + m.bottom() + p.top() + p.bottom()
        return layout.highestPage().zoomForHeight(height, layout.rotation, layout.dpiY)

    def pageSets(self, count):
        """Return a list of (count, length) tuples.

        Every count is the number of page sets of that length. When the layout
        is in non-continuous mode, it displays only a single page set at a time.
        For most layout engines, a page set is just one Page, but for column-
        based layouts other values make sense.

        """
        return [(count, 1)] if count else []


class RowLayoutEngine(LayoutEngine):
    """A layout engine that orders pages in rows.

    Additional instance attributes:

        `pagesPerRow`     = 2, the number of pages to display in a row
        `pagesFirstRow`   = 1, the number of pages to display in the first row
        `fitAllColumns`   = True, whether "fit width" uses all columns

    In non-continuous mode, this layout engine displayes a row of pages
    together. The `orientation` layout attribute is ignored in this layout
    engine.

    """

    pagesPerRow = 2
    pagesFirstRow = 1
    fitAllColumns = True

    orientation = Horizontal    # do not change

    def pageSets(self, count):
        """Return a list of (count, length) tuples respecting our column settings."""
        result = []
        left = count
        if left:
            if self.pagesFirstRow and self.pagesFirstRow != self.pagesPerRow:
                length = min(left, self.pagesFirstRow)
                result.append((1, length))
                left -= length
            if left:
                count, left = divmod(left, self.pagesPerRow)
                if count:
                    result.append((count, self.pagesPerRow))
                if left:
                    # merge result entries with same length
                    if result and result[-1][1] == left:
                        result[-1] == (result[-1][0] + 1, left)
                    else:
                        result.append((1, left))
        return result

    def grid(self, layout):
        """Return (ncols, nrows, prepend).

        Takes into account the pagesPerRow and pagesFirstRow instance
        variables. If desired, prepends empty positions so the first row
        contains less pages than the column width.

        """
        ncols = self.pagesPerRow
        if layout.count() > ncols:
            prepend = (ncols - self.pagesFirstRow) % ncols
        else:
            ncols = layout.count()
            prepend = 0
        nrows = math.ceil((layout.count() + prepend) / ncols)
        return ncols, nrows, prepend

    def zoomFitWidth(self, layout, width):
        """Reimplemented to respect the fitAllColumns setting."""
        if not self.fitAllColumns or self.pagesPerRow == 1 or layout.count() < 2:
            return super().zoomFitWidth(layout, width)
        ncols, nrows, prepend = self.grid(layout)
        m, p = layout.margins(), layout.pageMargins()
        width -= m.left() + m.right() + (p.left() + p.right()) * ncols
        width -= layout.spacing * (ncols - 1)
        if self.evenWidths:
            return super().zoomFitWidth(layout, width // ncols)
        # find the default width of the columns
        cols = [[] for n in range(ncols)]
        for page, (col, row) in self.pages(layout, ncols, nrows, prepend):
            cols[col].append(page)
        # widest page of every column
        widestpages = [max(col, key=layout.defaultWidth) for col in cols]
        totalDefaultWidth = sum(map(layout.defaultWidth, widestpages))
        return min(page.zoomForWidth(
                     width * layout.defaultWidth(page) // totalDefaultWidth,
                     layout.rotation, layout.dpiX)
            for page in widestpages)


class RasterLayoutEngine(LayoutEngine):
    """A layout engine that aligns the pages in a grid.

    This layout does not zoom to fit, but changes the number of columns and rows
    according to the available space. FitBoth is handled like FitWidth.

    """
    zoomToFit = False
    _h = 0
    _w = 0
    _mode = FixedScale

    def fit(self, layout, size, mode):
        """Reimplemented."""
        self._h = size.height()
        self._w = size.width()
        self._mode = mode

    def grid(self, layout):
        """Return a grid that would fit in the layout."""
        m, p = layout.margins(), layout.pageMargins()
        width = self._w - m.left() - m.right()
        height = self._h - m.top() - m.bottom()
        pmh = p.left() + p.right()        # horizontal page margin
        pmv = p.top() + p.bottom()        # vertical page margin

        if self._mode & FitWidth:
            w = layout.widestPage().width + pmh
            ncols = (width + layout.spacing) // (w + layout.spacing)
            if ncols:
                # this will fit, but try more
                for tryncols in range(ncols + 1, layout.count() + 1):
                    trynrows = math.ceil(layout.count() / tryncols)
                    cw, rh = self.dimensions(layout, tryncols, trynrows)
                    # compute width: column widths, spacing and page margins
                    w = sum(cw) + layout.spacing * (tryncols - 1) + pmh * tryncols
                    if w >= width:
                        ncols = tryncols - 1
                        break
                else:
                    ncols = layout.count()
            else:
                ncols = 1   # the minimum
            nrows = math.ceil(layout.count() / ncols)
        elif self._mode & FitHeight:
            h = layout.highestPage().height + pmv
            nrows = (height + layout.spacing) // (h + layout.spacing)
            if nrows:
                # this will fit, but try more
                for trynrows in range(nrows + 1, layout.count() + 1):
                    tryncols = math.ceil(layout.count() / trynrows)
                    cw, rh = self.dimensions(layout, tryncols, trynrows)
                    # compute height: row heights, spacing and page margins
                    h = sum(rh) + layout.spacing * (trynrows - 1) + pmv * trynrows
                    if h >= height:
                        nrows = trynrows - 1
                        break
                else:
                    nrows = layout.count()
            else:
                nrows = 1   # the minimum
            ncols = math.ceil(layout.count() / nrows)
        else:
            ncols = math.ceil(math.sqrt(layout.count()))
            nrows = math.ceil(layout.count() / ncols)
        return ncols, nrows, 0


# install a default layout engine at the class level
PageLayout.engine = LayoutEngine()




# -*- coding: utf-8 -*-
#
# This file is part of the qpageview package.
#
# Copyright (c) 2016 - 2020 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
Meta-information about the qpageview package.

This information is used by the install script.

"""

import collections
Version = collections.namedtuple("Version", "major minor patch")


#: name of the package
name = "qpageview"

#: the current version
version = Version(0, 6, 2)
version_suffix = ""
#: the current version as a string
version_string = "{}.{}.{}".format(*version) + version_suffix

#: short description
description = "Widget to display page-based documents for Qt5/PyQt5"

#: long description
long_description = \
    "The qpageview package provides a Python library to display page-based " \
    "documents, such as PDF and possibly other formats."

#: maintainer name
maintainer = "Wilbert Berendsen"

#: maintainer email
maintainer_email = "info@frescobaldi.org"

#: homepage
url = "https://github.com/frescobaldi/qpageview"

#: license
license = "GPL v3"

#: copyright year
copyright_year = "2020-2022"




# -*- coding: utf-8 -*-
#
# This file is part of the qpageview package.
#
# Copyright (c) 2019 - 2019 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
A MultiPage has no contents itself (but it has a size!), and renders a list of
embedded pages.

The MultiPageRenderer has the same interface as an ordinary renderer, but defers
rendering to the renderer of the embedded pages.

"""

import collections
import itertools

from PyQt5.QtCore import QPoint, QRect, QRectF, Qt
from PyQt5.QtGui import QColor, QImage, QPainter, QPixmap, QRegion, QTransform

from . import document
from . import page
from . import render



class MultiPage(page.AbstractRenderedPage):
    """A special Page that has a list of embedded sub pages.

    The sub pages are in the pages attribute, the first one is on top.

    The position and size of the embedded pages is set in the updateSize()
    method, which is inherited from AbstractPage. By default all sub pages
    are centered in their natural size.

    Rotation of sub pages is relative to the MultiPage.

    The `scalePages` instance attribute can be used to multiply the zoomfactor
    for the sub pages.

    The `opaquePages` instance attribute optimizes some procedures when set to
    True (i.e. it prevents rendering sub pages that are hidden below others).

    By default, only links in the first sub page are handled.
    Set `linksOnlyFirstSubPage` to False if you want links in all sub pages.

    """

    scalePages = 1.0
    opaquePages = True
    linksOnlyFirstSubPage = True

    def __init__(self, renderer=None):
        self.pages = []
        if renderer is not None:
            self.renderer = renderer

    @classmethod
    def createPages(cls, pageLists, renderer=None, pad=page.BlankPage):
        """Yield pages, taking each page from every pageList.

        If pad is given and is not None, it is a callable that instantiates
        blank pages, to pad the shorter pageLists with. In that case, the
        returned list of pages has the same length as the longest pageList
        given. If pad is None, the returned list of pages has the same length
        as the shortest pageList given.

        """
        it = itertools.zip_longest(*pageLists) if pad else zip(*pageLists)
        for pages in it:
            page = cls(renderer)
            page.pages[:] = (p if p else pad() for p in pages)
            yield page

    def copy(self, owner=None, matrix=None):
        """Reimplemented to also copy the sub pages."""
        page = super().copy(owner, matrix)
        page.pages = [p.copy(owner, matrix) for p in self.pages]
        return page

    def updateSize(self, dpiX, dpiY, zoomFactor):
        """Reimplemented to also position our sub-pages.

        The default implementation of this method zooms the sub pages
        at the zoom level of the page * self.scalePages.

        """
        super().updateSize(dpiX, dpiY, zoomFactor)

        # zoom the sub pages, using the same zoomFactor
        for page in self.pages:
            page.computedRotation = (page.rotation + self.computedRotation) & 3
            page.updateSize(dpiX, dpiY, zoomFactor * self.scalePages)

        self.updatePagePositions()

    def updatePagePositions(self):
        """Called by updateSize(), set the page positions.

        The default implementation of this method centers the pages.

        """
        # center the pages
        center = self.rect().center()
        for page in self.pages:
            r = page.rect()
            r.moveCenter(center)
            page.setGeometry(r)

    def visiblePagesAt(self, rect):
        """Yield (page, rect) for all subpages.

        The rect may be invalid when opaquePages is False. If opaquePages is
        True, pages outside rect or hidden below others are exclued. The
        yielded rect is always valid in that case.

        """
        if not self.opaquePages:
            for p in self.pages:
                yield p, rect & p.geometry()
        else:
            covered = QRegion()
            for p in self.pages:
                overlayrect = rect & p.geometry()
                if not overlayrect or not QRegion(overlayrect).subtracted(covered):
                    continue    # skip if this part is hidden below the other
                covered += overlayrect
                yield p, overlayrect
                if not QRegion(rect).subtracted(covered):
                    break

    def printablePagesAt(self, rect):
        """Yield (page, matrix) for all subpages that are visible in rect.

        If opaquePages is True, excludes pages outside rect or hidden below
        others. The matrix (QTransform) describes the transformation from the
        page to the sub page. Rect is in original coordinates, as with the
        print() method.

        """
        origmatrix = self.transform().inverted()[0] # map pos to original page
        origmatrix.scale(self.scaleX, self.scaleY)  # undo the scaling done in printing.py
        for p, r in self.visiblePagesAt(self.mapToPage().rect(rect)):
            center = origmatrix.map(QRectF(p.geometry()).center())
            m = QTransform()    # matrix from page to subpage
            m.translate(center.x(), center.y())
            m.rotate(p.rotation * 90) # rotation relative to us
            m.scale(
                self.scalePages * p.scaleX * self.dpi / p.dpi,
                self.scalePages * p.scaleY * self.dpi / p.dpi)
            m.translate(p.pageWidth / -2, p.pageHeight / -2)
            yield p, m

    def print(self, painter, rect=None, paperColor=None):
        """Prints our sub pages."""
        if rect is None:
            rect = self.pageRect()
        else:
            rect = rect & self.pageRect()
        painter.translate(-rect.topLeft())
        # print from bottom to top
        for p, m in reversed(list(self.printablePagesAt(rect))):
            # find center of the page corresponding to our center
            painter.save()
            painter.setTransform(m, True)
            # handle rect clipping
            clip = m.inverted()[0].mapRect(rect) & p.pageRect()
            painter.fillRect(clip, paperColor or Qt.white)    # draw a white background
            painter.translate(clip.topLeft())   # the page will go back...
            p.print(painter, clip)
            painter.restore()

    def text(self, rect):
        """Reimplemented to get text from sub pages."""
        for p, rect in self.visiblePagesAt(rect):
            if rect:
                text = p.text(rect.translated(-p.pos()))
                if text:
                    return text

    def _linkPages(self, rect=None):
        """Internal. Yield the pages allowed for links (and visible in rect if given)."""
        for p, rect in self.visiblePagesAt(rect or self.rect()):
            yield p, rect
            if self.linksOnlyFirstSubPage:
                break

    def linksAt(self, point):
        """Reimplemented to find links in sub pages."""
        result = []
        for p, rect in self._linkPages():
            if point in rect:
                result.extend(p.linksAt(point - p.pos()))
        return result

    def linksIn(self, rect):
        """Reimplemented to find links in sub pages."""
        result = set()
        for p, rect in self._linkPages(rect):
            result.update(p.linksIn(rect.translated(-p.pos())))
        return result

    def linkRect(self, link):
        """Reimplemented to get correct area on the page the link belongs to."""
        for p, r in self._linkPages():
            if link in p.links():
                return p.linkRect(link).translated(p.pos())
        return QRect()  # just in case


class MultiPageDocument(document.MultiSourceDocument):
    """A Document that combines pages from different documents."""
    pageClass = MultiPage
    def createPages(self):
        pageLists = [[p.copy() for p in doc.pages()] for doc in self.sources()]
        return self.pageClass.createPages(pageLists, self.renderer)


class MultiPageRenderer(render.AbstractRenderer):
    """A renderer that interfaces with the renderers of the sub pages of a MultiPage."""
    def update(self, page, device, rect, callback=None):
        """Reimplemented to check/rerender (if needed) all sub pages."""
        # make the call back return with the original page, not the overlay page
        newcallback = CallBack(callback, page) if callback else None

        ok = True
        for p, overlayrect in page.visiblePagesAt(rect):
            if (overlayrect and p.renderer and
                    not p.renderer.update(p, device, overlayrect.translated(-p.pos()), newcallback)):
                ok = False
        return ok

    def paint(self, page, painter, rect, callback=None):
        """Reimplemented to paint all the sub pages on top of each other."""
        # make the call back return with the original page, not the overlay page
        newcallback = CallBack(callback, page) if callback else None

        # get the device pixel ratio to paint for
        try:
            ratio = painter.device().devicePixelRatioF()
        except AttributeError:
            ratio = painter.device().devicePixelRatio()

        pixmaps = []
        covered = QRegion()
        for p, overlayrect in page.visiblePagesAt(rect):
            pixmap = QPixmap(overlayrect.size() * ratio)
            if not pixmap.isNull():
                pixmap.setDevicePixelRatio(ratio)
                pt = QPainter(pixmap)
                pt.translate(p.pos() - overlayrect.topLeft())
                p.paint(pt, overlayrect.translated(-p.pos()), newcallback)
                pt.end()
            # even an empty pixmap is appended, otherwise the layer count when
            # compositing goes awry
            pos = overlayrect.topLeft()
            pixmaps.append((pos, pixmap))
            covered += overlayrect

        if QRegion(rect).subtracted(covered):
            painter.fillRect(rect, page.paperColor or self.paperColor)

        self.combine(painter, pixmaps)

    def image(self, page, rect, dpiX, dpiY, paperColor):
        """Return a QImage of the specified rectangle, of all images combined."""

        overlays = []

        # find out the scale used for the image, to be able to position the
        # overlay images correctly (code copied from AbstractRenderer.image())
        s = page.defaultSize()
        hscale = s.width() * dpiX / page.dpi / page.width
        vscale = s.height() * dpiY / page.dpi / page.height
        ourscale = s.width() / page.width

        for p, overlayrect in page.visiblePagesAt(rect):
            # compute the correct resolution, find out which scale was
            # used by updateSize() (which may have been reimplemented)
            overlaywidth = p.pageWidth * p.scaleX * page.dpi / p.dpi
            if p.computedRotation & 1:
                overlayscale = overlaywidth / p.height
            else:
                overlayscale = overlaywidth / p.width
            scale = ourscale / overlayscale
            img = p.image(overlayrect.translated(-p.pos()), dpiX * scale, dpiY * scale, paperColor)
            pos = overlayrect.topLeft() - rect.topLeft()
            pos = QPoint(round(pos.x() * hscale), round(pos.y() * vscale))
            overlays.append((pos, img))

        image = QImage(rect.width() * hscale, rect.height() * vscale, self.imageFormat)
        image.fill(paperColor or page.paperColor or self.paperColor)
        self.combine(QPainter(image), overlays)
        return image

    def unschedule(self, pages, callback):
        """Reimplemented to unschedule all sub pages."""
        for page in pages:
            newcallback = CallBack(callback, page) if callback else None
            for p in page.pages:
                if p.renderer:
                    p.renderer.unschedule((p,), newcallback)

    def invalidate(self, pages):
        """Reimplemented to invalidate the base and overlay pages."""
        renderers = collections.defaultdict(list)
        for page in pages:
            for p in page.pages:
                if p.renderer:
                    renderers[p.renderer].append(p)
        for renderer, pages in renderers.items():
            renderer.invalidate(pages)

    def combine(self, painter, images):
        """Paints images on the painter.

        Each image is a tuple(QPoint, QPixmap), describing where to draw.
        The image on top is first, so drawing should start with the last.

        """
        for pos, image in reversed(images):
            if isinstance(image, QPixmap):
                painter.drawPixmap(pos, image)
            else:
                painter.drawImage(pos, image)


class CallBack:
    """A wrapper for a callable that is called with the original Page."""
    def __new__(cls, origcallable, page):
        # if the callable is already a CallBack instance, just return it. This
        # would happen if a MultiPage has a subpage that is also a MultiPage.
        if cls == type(origcallable):
            return origcallable
        cb = object.__new__(cls)
        cb.origcallable = origcallable
        cb.page = page
        return cb

    def __hash__(self):
        """Return the hash of the original callable.

        This way only one callback will be in the Job.callbacks attribute,
        despite of multiple pages, and unscheduling a job with subpages still
        works.

        """
        return hash(self.origcallable)

    def __call__(self, page):
        """Call the original callback with the original Page."""
        self.origcallable(self.page)



# install a default renderer, so MultiPage can be used directly
MultiPage.renderer = MultiPageRenderer()




# -*- coding: utf-8 -*-
#
# This file is part of the qpageview package.
#
# Copyright (c) 2016 - 2019 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
This is a generic paged view widget.

Its main design goal is to display the pages of a PDF document, but it can
display any set of images or pages, originating from different documents.

Every page is represented by a Page instance, which encompasses all logic
for the document type, i.e. drawing etc.

Pages are managed by a PageLayout.

A PageLayout can be set to a View so the pages are displayed.

The images from a PDF, SVG of possibly other document are cached, and rendering
is tile-based, to support zooming in at great detail.  Also a magnifier is
available, which by default pops up at Ctrl+click.

Because the qpageview is built on Qt, we use the Qt convention
to have camelCase method names and CamelCase class names.

"""

from .constants import (
    # rotation:
    Rotate_0,
    Rotate_90,
    Rotate_180,
    Rotate_270,


    # viewModes:
    FixedScale,
    FitWidth,
    FitHeight,
    FitBoth,


    # orientation:
    Horizontal,
    Vertical,

)

from . import link
from . import highlight
from . import shadow
from . import view
from . import util
from .pkginfo import version, version_string



class View(
        link.LinkViewMixin,
        highlight.HighlightViewMixin,
        shadow.ShadowViewMixin,
        util.LongMousePressMixin,
        view.View,
    ):
    """Paged view component based on view.View, with all enhancements."""
    pass


def loadPdf(filename, renderer=None):
    """Convenience function to create a Document with the specified PDF file.

    The filename can also be a QByteArray or an already loaded
    popplerqt5.Poppler.Document instance.

    """
    from . import poppler
    return poppler.PopplerDocument(filename, renderer)


def loadSvgs(filenames, renderer=None):
    """Convenience function to create a Document with the specified SVG files.

    Each SVG file is loaded in one Page. A filename can also be a
    QByteArray.

    """
    from . import svg
    return svg.SvgDocument(filenames, renderer)


def loadImages(filenames, renderer=None):
    """Convenience function to create a Document with images from files.

    Each image is loaded in one Page. A filename can also be a
    QByteArray.

    """
    from . import image
    return image.ImageDocument(filenames, renderer)


def diffDocument(documents, renderer=None):
    """Convenience function to create a diff.DiffDocument from other documents.

    The difference works best with documents that are similar and mostly black
    and white.

    """
    from . import diff
    return diff.DiffDocument(documents, renderer)





# -*- coding: utf-8 -*-
#
# This file is part of the qpageview package.
#
# Copyright (c) 2019 - 2019 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
Small utilities and simple base classes for the qpageview module.
"""


import collections
import contextlib

from PyQt5.QtCore import QPoint, QPointF, QRect, QRectF, QSize, Qt
from PyQt5.QtGui import QBitmap, QMouseEvent, QRegion
from PyQt5.QtWidgets import QApplication


class Rectangular:
    """Defines a Qt-inspired and -based interface for rectangular objects.

    The attributes x, y, width and height default to 0 at the class level
    and can be set and read directly.

    For convenience, Qt-styled methods are available to access and modify these
    attributes.

    """
    x = 0
    y = 0
    width = 0
    height = 0

    def setPos(self, point):
        """Set the x and y coordinates from the given QPoint point."""
        self.x = point.x()
        self.y = point.y()

    def pos(self):
        """Return our x and y coordinates as a QPoint(x, y)."""
        return QPoint(self.x, self.y)

    def setSize(self, size):
        """Set the height and width attributes from the given QSize size."""
        self.width = size.width()
        self.height = size.height()

    def size(self):
        """Return the height and width attributes as a QSize(width, height)."""
        return QSize(self.width, self.height)

    def setGeometry(self, rect):
        """Set our x, y, width and height directly from the given QRect."""
        self.x, self.y, self.width, self.height = rect.getRect()

    def geometry(self):
        """Return our x, y, width and height as a QRect."""
        return QRect(self.x, self.y, self.width, self.height)

    def rect(self):
        """Return QRect(0, 0, width, height)."""
        return QRect(0, 0, self.width, self.height)


class MapToPage:
    """Simple class wrapping a QTransform to map rect and point to page coordinates."""
    def __init__(self, transform):
        self.t = transform

    def rect(self, rect):
        """Convert QRect or QRectF to a QRect in page coordinates."""
        return self.t.mapRect(QRectF(rect)).toRect()

    def point(self, point):
        """Convert QPointF or QPoint to a QPoint in page coordinates."""
        return self.t.map(QPointF(point)).toPoint()


class MapFromPage(MapToPage):
    """Simple class wrapping a QTransform to map rect and point from page to original coordinates."""
    def rect(self, rect):
        """Convert QRect or QRectF to a QRectF in original coordinates."""
        return self.t.mapRect(QRectF(rect))

    def point(self, point):
        """Convert QPointF or QPoint to a QPointF in original coordinates."""
        return self.t.map(QPointF(point))


class LongMousePressMixin:
    """Mixin class to add support for long mouse press to a QWidget.

    To handle a long mouse press event, implement longMousePressEvent().

    """

    #: Whether to enable handling of long mouse presses; set to False to disable
    longMousePressEnabled = True

    #: Allow moving some pixels before a long mouse press is considered a drag
    longMousePressTolerance = 3

    #: How long to presse a mouse button (in msec) for a long press
    longMousePressTime = 800

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._longPressTimer = None
        self._longPressAttrs = None
        self._longPressPos = None

    def _startLongMousePressEvent(self, ev):
        """Start the timer for a QMouseEvent mouse press event."""
        self._cancelLongMousePressEvent()
        self._longPressTimer = self.startTimer(self.longMousePressTime)
        # copy the event's attributes because Qt might reuse the event
        self._longPressAttrs = (ev.type(),
            ev.localPos(), ev.windowPos(), ev.screenPos(),
            ev.button(), ev.buttons(), ev.modifiers())
        self._longPressPos = ev.pos()

    def _checkLongMousePressEvent(self, ev):
        """Cancel the press event if the current event has moved more than 3 pixels."""
        if self._longPressTimer is not None:
            dist = (self._longPressPos - ev.pos()).manhattanLength()
            if dist > self.longMousePressTolerance:
                self._cancelLongMousePressEvent()

    def _cancelLongMousePressEvent(self):
        """Stop the timer for a long mouse press event."""
        if self._longPressTimer is not None:
            self.killTimer(self._longPressTimer)
            self._longPressTimer = None
            self._longPressAttrs = None
            self._longPressPos = None

    def longMousePressEvent(self, ev):
        """Implement this to handle a long mouse press event."""
        pass

    def timerEvent(self, ev):
        """Implemented to check for a long mouse button press."""
        if ev.timerId() == self._longPressTimer:
            event = QMouseEvent(*self._longPressAttrs)
            self._cancelLongMousePressEvent()
            self.longMousePressEvent(event)
        super().timerEvent(ev)

    def mousePressEvent(self, ev):
        """Reimplemented to check for a long mouse button press."""
        if self.longMousePressEnabled:
            self._startLongMousePressEvent(ev)
        super().mousePressEvent(ev)

    def mouseMoveEvent(self, ev):
        """Reimplemented to check for moves during a long press."""
        self._checkLongMousePressEvent(ev)
        super().mouseMoveEvent(ev)

    def mouseReleaseEvent(self, ev):
        """Reimplemented to cancel a long press."""
        self._cancelLongMousePressEvent()
        super().mouseReleaseEvent(ev)


def rotate(matrix, rotation, width, height, dest=False):
    """Rotate matrix inside a rectangular area of width x height.

    The ``matrix`` can be a either a QPainter or a QTransform. The ``rotation``
    is 0, 1, 2 or 3, etc. (``Rotate_0``, ``Rotate_90``, etc...). If ``dest`` is
    True, ``width`` and ``height`` refer to the destination, otherwise to the
    source.

    """
    if rotation & 3:
        if dest or not rotation & 1:
            matrix.translate(width / 2, height / 2)
        else:
            matrix.translate(height / 2, width / 2)
        matrix.rotate(rotation * 90)
        if not dest or not rotation & 1:
            matrix.translate(width / -2, height / -2)
        else:
            matrix.translate(height / -2, width / -2)


def align(w, h, ow, oh, alignment=Qt.AlignCenter):
    """Return (x, y) to align a rect w x h in an outer rectangle ow x oh.

    The alignment can be a combination of the Qt.Alignment flags.
    If w > ow, x = -1; and if h > oh, y = -1.

    """
    if w > ow:
        x = -1
    elif alignment & Qt.AlignHCenter:
        x = (ow - w) // 2
    elif alignment & Qt.AlignRight:
        x = ow - w
    else:
        x = 0
    if h > oh:
        y = -1
    elif alignment & Qt.AlignVCenter:
        y = (oh - h) // 2
    elif alignment & Qt.AlignBottom:
        y = oh - h
    else:
        y = 0
    return x, y


def alignrect(rect, point, alignment=Qt.AlignCenter):
    """Align rect with point according to the alignment.

    The alignment can be a combination of the Qt.Alignment flags.

    """
    rect.moveCenter(point)
    if alignment & Qt.AlignLeft:
        rect.moveLeft(point.x())
    elif alignment & Qt.AlignRight:
        rect.moveRight(point.x())
    if alignment & Qt.AlignTop:
        rect.moveTop(point.y())
    elif alignment & Qt.AlignBottom:
        rect.moveBottom(point.y())


# Found at: https://stackoverflow.com/questions/1986152/why-doesnt-python-have-a-sign-function
def sign(x):
    """Return the sign of x: -1 if x < 0, 0 if x == 0, or 1 if x > 0."""
    return bool(x > 0) - bool(x < 0)


@contextlib.contextmanager
def signalsBlocked(*objs):
    """Block the pyqtSignals of the given QObjects during the context."""
    blocks = [obj.blockSignals(True) for obj in objs]
    try:
        yield
    finally:
        for obj, block in zip(objs, blocks):
            obj.blockSignals(block)


def autoCropRect(image):
    """Return a QRect specifying the contents of the QImage.

    Edges of the image are trimmed if they have the same color.

    """
    # pick the color at most of the corners
    colors = collections.defaultdict(int)
    w, h = image.width(), image.height()
    for x, y in (0, 0), (w - 1, 0), (w - 1, h - 1), (0, h - 1):
        colors[image.pixel(x, y)] += 1
    most = max(colors, key=colors.get)
    # let Qt do the masking work
    mask = image.createMaskFromColor(most)
    return QRegion(QBitmap.fromImage(mask)).boundingRect()


def tempdir():
    """Return a temporary directory that is erased on app quit."""
    import tempfile
    global _tempdir
    try:
        _tempdir
    except NameError:
        name = QApplication.applicationName().translate({ord('/'): None}) or 'qpageview'
        _tempdir = tempfile.mkdtemp(prefix=name + '-')
        import atexit
        import shutil
        @atexit.register
        def remove():
            shutil.rmtree(_tempdir, ignore_errors=True)
    return tempfile.mkdtemp(dir=_tempdir)




# -*- coding: utf-8 -*-
#
# This file is part of the qpageview package.
#
# Copyright (c) 2016 - 2019 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
Infrastructure for rendering and caching Page images.
"""

import collections
import sys
import time

from PyQt5.QtCore import QRect, QRectF, Qt
from PyQt5.QtGui import QColor, QImage, QPainter, QRegion, QTransform

from . import backgroundjob
from . import cache
from . import util

#: Describes a tile to render. Most times all coordinates are integers.
#: The needed tiles for a page are yielded by :meth:`AbstractRenderer.tiles`.
Tile = collections.namedtuple('Tile', 'x y w h')
Tile.x.__doc__ = "The x coordinate of the tile"
Tile.y.__doc__ = "The y coordinate of the tile"
Tile.w.__doc__ = "The width of the tile"
Tile.h.__doc__ = "The height of the tile"

#: Identifies a render operation for a Page, returned by
#: :meth:`AbstractRenderer.key`.
Key = collections.namedtuple("Key", "group ident rotation width height")
Key.group.__doc__ = "The :meth:`~.page.AbstractPage.group` of the page"
Key.ident.__doc__ = "The :meth:`~.page.AbstractPage.ident` of the page"
Key.rotation.__doc__ = "The :attr:`~.page.AbstractPage.computedRotation` of the page"
Key.width.__doc__ = "The :attr:`~.util.Rectangular.width` of the page"
Key.height.__doc__ = "The :attr:`~.util.Rectangular.height` of the page"

#: Information about cached or missing rendered tiles to display a rectangular
#: part of a Page at a certain size. Returned by :meth:`AbstractRenderer.info`.
RenderInfo = collections.namedtuple("RenderInfo", "images missing key target ratio")
RenderInfo.images.__doc__ = "a list of tuples (tile, image) that are available in the cache"
RenderInfo.missing.__doc__ = "a list of Tile instances that are needed but not available in the cache"
RenderInfo.key.__doc__ = "the Key returned by :meth:`~AbstractRenderer.key`, describing width, height, rotation and identity of the page"
RenderInfo.target.__doc__ = "the rect multiplied by the ratio"
RenderInfo.images.__doc__ = "the devicepixelratio of the specified paint device"

# the maximum number of concurrent jobs (at global level)
maxjobs = 4

# we use a global dict to keep running jobs in, so a thread is never
# deallocated when a renderer dies.
_jobs = {}



class AbstractRenderer:
    """Handle rendering and caching of images.

    A renderer can be assigned to the renderer attribute of a Page and takes
    care for generating, caching and updating the images needed for display
    of the Page at different sizes.

    You can use a renderer for as many Page instances as you like. You can use
    one global renderer in your application or more, depending on how you use
    the qpageview package.

    You must inherit from this class and at least implement the render() or the
    draw() method.

    Instance attributes:

    ``paperColor``
        Paper color. If possible this background color is used when rendering
        the pages, also for temporary drawings when a page has to be rendered.
        If a Page specifies its own paperColor, that color prevails.

    ``imageFormat``
        QImage format to use (if possible). Default is
        QImage.Format_ARGB32_Premultiplied

    ``antialiasing``
        True by default. Whether to antialias graphics. (Most Renderers
        antialias anyway, even if this is False.)

    """

    MAX_TILE_WIDTH = 2400
    MAX_TILE_HEIGHT = 1600

    # default paper color to use (if possible, and when drawing an empty page)
    paperColor = QColor(Qt.white)

    # QImage format to use (if possible)
    imageFormat = QImage.Format_ARGB32_Premultiplied

    # antialias True by default (not all renderers may support this)
    antialiasing = True

    def __init__(self, cache=None):
        if cache:
            self.cache = cache

    def copy(self):
        """Return a copy of the renderer, with always a new cache."""
        c = self.cache
        if c:
            c = type(c)()
            c.__dict__.update(self.cache.__dict__)
        r = type(self)(c)
        r.__dict__.update(self.__dict__)
        return r

    @staticmethod
    def key(page, ratio):
        """Return a five-tuple Key describing the page.

        The ratio is a device pixel ratio; width and height are multiplied
        with this value, to render and cache an image correctly on high-
        density displays.

        This is used for rendering and caching. It is never stored as is.
        The cache can store the group object using a weak reference.
        The tuple contains the following values:

        ``group``
            the object returned by ``page.group()``
        ``ident``
            the value returned by ``page.ident()``
        ``rotation``
            ``page.computedRotation``
        ``width``
            ``page.width * ratio``
        ``height``
            ``page.height * ratio``

        """
        return Key(
            page.group(),
            page.ident(),
            page.computedRotation,
            int(page.width * ratio),
            int(page.height * ratio),
        )

    def tiles(self, width, height):
        """Yield four-tuples Tile(x, y, w, h) describing the tiles to render."""
        rowcount = height // self.MAX_TILE_HEIGHT
        colcount = width  // self.MAX_TILE_WIDTH
        tilewidth, extrawidth = divmod(width, colcount + 1)
        tileheight, extraheight = divmod(height, rowcount + 1)
        rows = [tileheight] * rowcount + [tileheight + extraheight]
        cols = [tilewidth] * colcount + [tilewidth + extrawidth]
        y = 0
        for h in rows:
            x = 0
            for w in cols:
                yield Tile(x, y, w, h)
                x += w
            y += h

    def map(self, key, box):
        """Return a QTransform converting from Key coordinates to a box.

        The box should be a QRectF or QRect, and describes the original area of
        the page.  The returned matrix can be used to convert e.g. tile
        coordinates to the position on the original page.

        """
        t = QTransform()
        t.translate(box.x(), box.y())
        t.scale(box.width(), box.height())
        t.translate(.5, .5)
        t.rotate(-key.rotation * 90)
        t.translate(-.5, -.5)
        t.scale(1 / key.width, 1 / key.height)
        return t

    def image(self, page, rect, dpiX, dpiY, paperColor):
        """Returns a QImage of the specified rectangle on the Page.

        The rectangle is relative to the top-left position. The image is not
        cached.

        """
        s = page.defaultSize()
        hscale = s.width() * dpiX / page.dpi / page.width
        vscale = s.height() * dpiY / page.dpi / page.height
        matrix = QTransform().scale(hscale, vscale)

        tile = Tile(*matrix.mapRect(rect).getRect())
        key = Key(page.group(),
                  page.ident(),
                  page.computedRotation,
                 *matrix.map(page.width, page.height))
        return self.render(page, key, tile, paperColor)

    def render(self, page, key, tile, paperColor=None):
        """Generate a QImage for tile of the Page.

        The width, height and rotation to render at should be taken from the
        key, as the page could be resized or rotated in the mean time.

        The default implementation prepares the image, a painter and then
        calls draw() to actually draw the contents.

        If the paperColor is not specified, it will be read from the Page's
        paperColor attribute (if not None) or else from the renderer's
        paperColor attribute.

        """
        if paperColor is None:
            paperColor = page.paperColor or self.paperColor

        i = QImage(tile.w, tile.h, self.imageFormat)
        i.fill(paperColor)
        painter = QPainter(i)

        # rotate the painter accordingly
        util.rotate(painter, key.rotation, tile.w, tile.h, True)

        # draw it on the image
        self.draw(page, painter, key, tile, paperColor)
        return i

    def draw(self, page, painter, key, tile, paperColor=None):
        """Draw the page contents; implement at least this method.

        The painter is already at the top-left position and the correct
        rotation. You should convert the tile to the original area on the page,
        you can use the map() method for that. You can draw in tile/key
        coordinates. Don't use width, height and rotation from the Page object,
        as it could have been resized or rotated in the mean time.

        The paperColor can be speficied, but it is not needed to paint it: by
        default the render() method already fills the image, and when drawing on
        a printer, painting the background is normally not desired.

        """
        pass

    def info(self, page, device, rect):
        """Return a namedtuple RenderInfo(images, missing, key, target, ratio).

        images is a list of tuples (tile, image) that are available in the
        cache; missing is a list of Tile instances that are not available in
        the cache; key is the Key returned by key(), describing width, height,
        rotation and identity of the page; target is the rect multiplied by the
        ratio; which is the devicepixelratio of the specified paint device.

        """
        try:
            ratio = device.devicePixelRatioF()
        except AttributeError:
            ratio = device.devicePixelRatio()
        key = self.key(page, ratio)

        # paint rect in tile coordinates
        target = QRect(int(rect.x() * ratio), int(rect.y() * ratio), int(rect.width() * ratio), int(rect.height() * ratio))

        # tiles to paint
        tiles = [t for t in self.tiles(key.width, key.height) if QRect(*t) & target]

        # look in cache, get a dict with tiles and their images
        tileset = self.cache.tileset(key)

        images = []
        missing = []
        for t in tiles:
            entry = tileset.get(t)
            if entry:
                entry.time = time.time()    # prevent aging ;-)
                images.append((t, entry.image))
            else:
                missing.append(t)

        return RenderInfo(images, missing, key, target, ratio)

    def update(self, page, device, rect, callback=None):
        """Check if a page can be painted on the device without waiting.

        Return True if that is the case. Otherwise schedules missing tiles
        for rendering and calls the callback each time one tile if finished.

        """
        info = self.info(page, device, rect)
        if info.missing:
            self.schedule(page, info.key, info.missing, callback)
            return False
        return True

    def paint(self, page, painter, rect, callback=None):
        """Paint a page, using images from the cache.

        ``page``:
            the Page to draw

        ``painter``:
            the QPainter to use to draw

        ``rect``:
            the region to draw, relative to the topleft of the page.

        ``callback``:
            if specified, a callable accepting the `page` argument.
            Typically this should be used to trigger a repaint of the view.

        The Page calls this method by default in its
        :meth:`~.page.AbstractPage.paint` method. This method tries to fetch an
        image from the cache and paint that. If no image is available, render()
        is called in the background to generate one. If it is ready, the
        callback is called with the Page as argument. An interim image may be
        painted in the meantime (e.g. scaled from another size).

        """
        images = [] # list of images to draw at end of this method
        region = QRegion() # painted region in tile coordinates

        info = self.info(page, painter.device(), rect)

        for t, image in info.images:
            r = QRect(*t) & info.target # part of the tile that needs to be drawn
            images.append((r, image,  QRectF(r.translated(-t.x, -t.y))))
            region += r

        if info.missing:
            self.schedule(page, info.key, info.missing, callback)

            # find other images from cache for missing tiles
            for width, height, tileset in self.cache.closest(info.key):
                # we have a dict of tiles for an image of size width x height
                hscale = info.key.width / width
                vscale = info.key.height / height
                for t in tileset:
                    # scale to our image size
                    r = QRect(int(t.x * hscale), int(t.y * vscale), int(t.w * hscale), int(t.h * vscale)) & info.target
                    if r and QRegion(r).subtracted(region):
                        # we have an image that can be drawn in rect r
                        source = QRectF(r.x() / hscale - t.x, r.y() / vscale - t.y,
                                        r.width() / hscale, r.height() / vscale)
                        images.append((r, tileset[t].image, source))
                        region += r
                        # stop if we have covered the whole drawing area
                        if not QRegion(info.target).subtracted(region):
                            break
                else:
                    continue
                break
            else:
                if QRegion(info.target).subtracted(region):
                    # paint background, still partly uncovered
                    painter.fillRect(rect, page.paperColor or self.paperColor)

        # draw lowest quality images first
        for (r, image, source) in reversed(images):
            # scale the target rect back to the paint device
            target = QRectF(r.x() / info.ratio, r.y() / info.ratio, r.width() / info.ratio, r.height() / info.ratio)
            painter.drawImage(target, image, source)

    def schedule(self, page, key, tiles, callback):
        """Schedule a new rendering job for the specified tiles of the page.

        If this page has already a job pending, the callback is added to the
        pending job.

        """
        for tile in tiles:
            try:
                job = _jobs[(key, tile)]
            except KeyError:
                # make a new Job for this tile
                job = _jobs[(key, tile)] = self.job(page, key, tile)
            job.time = time.time()
            job.callbacks.add(callback)
        self.checkstart()

    def job(self, page, key, tile):
        """Return a new :class:`~.backgroundjob.Job` tailored for this tile."""
        job = backgroundjob.Job()
        job.callbacks = callbacks = set()
        job.mutex = page.mutex()
        exception = []
        def work():
            try:
                return self.render(page, key, tile)
            except Exception:
                exception.extend(sys.exc_info())
                return QImage()
        def finalize(image):
            self.cache.addtile(key, tile, image)
            for cb in callbacks:
                cb(page)
            del _jobs[(key, tile)]
            self.checkstart()
            if exception:
                self.exception(*exception)
        job.work = work
        job.finalize = finalize
        return job

    def unschedule(self, pages, callback):
        """Unschedule a possible pending rendering job for the given pages.

        If the pending job has no other callbacks left, it is removed,
        unless it is running.

        """
        pages = set((p.group(), p.ident()) for p in pages)
        unschedule = []
        for (key, tile), job in _jobs.items():
            if key[:2] in pages:
                job.callbacks.discard(callback)
                if not job.callbacks and not job.running:
                    unschedule.append((key, tile))
        for jobkey in unschedule:
            job = _jobs.pop(jobkey)
            job.finalize = job.work = None

    def invalidate(self, pages):
        """Delete the cached images for the given pages."""
        for p in pages:
            self.cache.invalidate(p)

    def checkstart(self):
        """Check whether there are jobs that need to be started.

        This method is called by the schedule() method, and by the finish()
        method when a job finishes, so that the number of running jobs never
        exceeds `maxjobs`.

        """
        runningjobs = [job for job in _jobs.values() if job.running]
        waitingjobs = sorted((job for job in _jobs.values() if not job.running),
                        key=lambda j: j.time, reverse=True)     # newest first

        jobcount = maxjobs - len(runningjobs)
        if jobcount > 0:
            mutexes = set(j.mutex for j in runningjobs)
            mutexes.discard(None)
            for job in waitingjobs:
                m = job.mutex
                if m is None or m not in mutexes:
                    mutexes.add(m)
                    job.start()
                    jobcount -= 1
                    if jobcount == 0:
                        break

    def exception(self, exctype, excvalue, exctb):
        """Called when an exception has occurred in a background rendering job.

        The default implementation prints a traceback to stderr.

        """
        import traceback
        traceback.print_exception(exctype, excvalue, exctb)



# install a global cache to use by default
AbstractRenderer.cache = cache.ImageCache()





# -*- coding: utf-8 -*-
#
# This file is part of the qpageview package.
#
# Copyright (c) 2019 - 2019 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
A Page intended to display the visual difference between other pages.

"""

import itertools

from PyQt5.QtCore import Qt
from PyQt5.QtGui import QColor, QPainter, QPixmap

from . import multipage
from . import page


class DiffPage(page.ImagePrintPageMixin, multipage.MultiPage):
    """A Page that shows the difference between sub pages.

    DiffPage inherits from MultiPage; the pages are to be added in the pages
    attribute. The first page is considered to be the "default" page, shown
    the normal way; the others are added in configurable colors and intensity.

    """
    opaquePages = False

    @classmethod
    def createPages(cls, pageLists, renderer=None, pad=page.BlankPage):
        """Reimplemented to adapt the page sizes."""
        it = itertools.zip_longest(*pageLists) if pad else zip(*pageLists)
        for pages in it:
            page = cls(renderer)
            # copy the dimensions from the first non-blank page
            for p in pages:
                if p:
                    page.dpi = p.dpi
                    page.pageWidth = p.pageWidth
                    page.pageHeight = p.pageHeight
                    break
            # set that dimensions also to blank pages.
            def padpage():
                p = pad()
                p.dpi = page.dpi
                p.pageWidth = page.pageWidth
                p.pageHeight = page.pageHeight
                return p
            page.pages[:] = (p if p else padpage() for p in pages)
            yield page


class DiffDocument(multipage.MultiPageDocument):
    """A Document showing the differences between documents, set as sources."""
    pageClass = DiffPage


class DiffRenderer(multipage.MultiPageRenderer):
    """Renders the pages by calling their own renderer.

    How the difference is displayed can be configured using this renderer.
    Up to four different pages can be displayed, the colors to render them
    are taken from the colors instance variable, which is a list.

    The alpha channel of each color determines the visiblity of the
    corresponding sub page.

    This renderer works best with pages that are mostly black on a white
    background.

    """
    def __init__(self):
        # we don't use a cache so no need to call super init
        self.colors = [
            QColor(Qt.black),
            QColor(Qt.red),
            QColor(Qt.green),
            QColor(Qt.blue),
        ]

    def combine(self, painter, images):
        """Paint images on the painter.

        We draw bottom-up, using Darken composition mode, so the lower images
        remain visible.

        """
        for color, (pos, image) in zip(self.colors, images):
            # take the alpha component
            intensity = 255 - color.alpha()
            if intensity == 255:
                continue    # the image would appear white anyway
            color = color.rgb()
            color |= 0x010101 * intensity
            color |= 0xFF000000

            p = QPainter(image)
            p.setCompositionMode(QPainter.CompositionMode_Lighten)
            p.fillRect(image.rect(), QColor(color))
            p.end()
            if isinstance(image, QPixmap):
                painter.drawPixmap(pos, image)
            else:
                painter.drawImage(pos, image)
            painter.setCompositionMode(QPainter.CompositionMode_Darken)



# install a default renderer, so DiffPage can be used directly
DiffPage.renderer = DiffRenderer()






# -*- coding: utf-8 -*-
#
# This file is part of the qpageview package.
#
# Copyright (c) 2016 - 2019 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
Cache logic.
"""

import weakref
import time


class ImageEntry:
    def __init__(self, image):
        self.image = image
        self.bcount = image.byteCount()
        self.time = time.time()


class ImageCache:
    """Cache generated images.

    Store and retrieve them under a key (see render.Renderer.key()).

    """
    maxsize = 209715200 # 200M
    currentsize = 0

    def __init__(self):
        self._cache = weakref.WeakKeyDictionary()

    def clear(self):
        """Remove all cached images."""
        self._cache.clear()
        self.currentsize = 0

    def invalidate(self, page):
        """Clear cache contents for the specified page."""
        try:
            del self._cache[page.group()][page.ident()]
        except KeyError:
            pass

    def tileset(self, key):
        """Return a dictionary with tile-entry pairs for the key.

        If no single tile is available, an empty dict is returned.

        """
        try:
            return self._cache[key.group][key.ident][key[2:]]
        except KeyError:
            return {}

    def addtile(self, key, tile, image):
        """Add image for the specified key and tile."""
        d = self._cache.setdefault(key.group, {}).setdefault(key.ident, {}).setdefault(key[2:], {})
        try:
            self.currentsize -= d[tile].bcount
        except KeyError:
            pass

        purgeneeded = self.currentsize > self.maxsize

        e = d[tile] = ImageEntry(image)
        self.currentsize += e.bcount

        if not purgeneeded:
            return

        # purge old images is needed,
        # cache groups may have disappeared so count all images

        entries = iter(sorted(
            ((entry.time, entry.bcount, group, ident, key, tile)
            for group, identd in self._cache.items()
                for ident, keyd in identd.items()
                    for key, tiled in keyd.items()
                        for tile, entry in tiled.items()),
            key=(lambda item: item[:2]), reverse=True))

        # now count the newest images until maxsize ...
        currentsize = 0
        for time, bcount, group, ident, key, tile in entries:
            currentsize += bcount
            if currentsize > self.maxsize:
                break
        self.currentsize = currentsize
        # ... and delete the remaining images, deleting empty dicts as well
        for time, bcount, group, ident, key, tile in entries:
            del self._cache[group][ident][key][tile]
            if not self._cache[group][ident][key]:
                del self._cache[group][ident][key]
                if not self._cache[group][ident]:
                    del self._cache[group][ident]
                    if not self._cache[group]:
                        del self._cache[group]

    def closest(self, key):
        """Iterate over suitable image tilesets but with a different size.

        Yields (width, height, tileset) tuples.

        This can be used for interim display while the real image is being
        rendered.

        """
        # group and ident must be there.
        try:
            keyd = self._cache[key.group][key.ident]
        except KeyError:
            return ()

        # prevent returning images that are too small
        minwidth = min(100, key.width / 2)

        suitable = [
            (k[1], k[2], tileset)
            for k, tileset in keyd.items()
                if k[0] == key.rotation and k[1] != key.width and k[1] > minwidth]
        return sorted(suitable, key=lambda s: abs(1 - s[0] / key.width))





# -*- coding: utf-8 -*-
#
# This file is part of the qpageview package.
#
# Copyright (c) 2019 - 2019 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
ImageView, a View optimized for display of one Page, e.g. one image.

Clicking in the view toggles between FitBoth and NaturalSize.

"""


from PyQt5.QtCore import QMargins, Qt

from . import constants
from . import util
from . import view


class ImageViewMixin:
    """View Mixin with a few customisations for displaying a single page/image.

    Adds the instance variable:

        fitNaturalSizeEnabled = True

                If True, the image will not be scaled larger than its natural
                size when FitWidth, -Height, or -Both is active.

    """
    fitNaturalSizeEnabled = True

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setViewMode(constants.FitBoth)
        self.pageLayout().setMargins(QMargins(0, 0, 0, 0))

    def setImage(self, image):
        """Convenience method to display a QImage."""
        self.loadImages([image])

    def toggleZooming(self):
        """Toggles between FitBoth and natural size."""
        if self.viewMode() == constants.FitBoth:
            self.setViewMode(constants.FixedScale)
            self.zoomNaturalSize()
        else:
            self.setViewMode(constants.FitBoth)

    def fitPageLayout(self):
        """Reimplemented to avoid zooming-to-fit larger than naturalsize."""
        layout = self.pageLayout()
        if self.fitNaturalSizeEnabled and self.viewMode() and layout.count():
            zoom_factor = layout.zoomFactor
            # fit layout but prevent zoomFactorChanged from being emitted
            with util.signalsBlocked(self):
                super().fitPageLayout()
            # what would be the natural size?
            factor = layout[0].dpi / self.physicalDpiX()
            # adjust if the image was scaled larger
            if layout.zoomFactor > factor:
                layout.zoomFactor = factor
            if zoom_factor != layout.zoomFactor:
                self.zoomFactorChanged.emit(layout.zoomFactor)
        else:
            super().fitPageLayout()

    def mouseReleaseEvent(self, ev):
        """Reimplemented to toggle between FitBoth and ZoomNaturalSize."""
        if not self.isDragging() and ev.button() == Qt.LeftButton:
            self.toggleZooming()
        super().mouseReleaseEvent(ev)


class ImageView(ImageViewMixin, view.View):
    """A View, optimized for display of one Page, e.g. one image.

    Append one Page to the layout, use one of the load* methods to load
    a single page document, or use the setImage() method to display a QImage.

    """
    clickToSetCurrentPageEnabled = False





# -*- coding: utf-8 -*-
#
# This file is part of the qpageview package.
#
# Copyright (c) 2019 - 2019 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
A page that can display an image, loaded using QImage.

ImagePages are instantiated quite fast. The image is only really loaded on first
display.

"""

from PyQt5.QtCore import QPoint, QRect, QSize, Qt
from PyQt5.QtGui import QImage, QImageIOHandler, QImageReader, QPainter, QTransform

from . import document
from . import locking
from . import page
from . import render


class ImageContainer:
    """Represent an image, is shared among copies of the "same" Page."""
    def __init__(self, image):
        """Init with a QImage."""
        self._image = image

    def size(self):
        return self._image.size()

    def image(self, clip=None):
        if clip is None:
            return self._image
        return self._image.copy(clip)


class ImageLoader(ImageContainer):
    """Represent an image loaded from a file or IO device."""
    def __init__(self, source, autoTransform=True):
        """Init with a filename or QIODevice.

        If autoTransform is True (the default), EXIF rotation is automatically
        applied when loading the image.

        """
        self._size = None
        self.source = source
        self.autoTransform = autoTransform

    def _reader(self):
        """Return a QImageReader for the source."""
        reader = QImageReader(self.source)
        reader.setAutoTransform(self.autoTransform)
        return reader

    def size(self):
        """Return the size of the image.

        If the image can't be loaded, a null size is returned. The resulting
        value is cached.

        """
        if self._size is None:
            self._size = QSize()
            reader = self._reader()
            if reader.canRead():
                size = reader.size()
                if size:
                    if self.autoTransform and reader.transformation() & 4:
                        size.transpose()
                    self._size = size
        return QSize(self._size)

    def image(self, clip=None):
        """Load and return the image.

        If clip is given, it should be a QRect describing the area to load.

        """
        with locking.lock(self):
            reader = self._reader()
            if clip:
                if self.autoTransform:
                    size = reader.size()
                    transf = reader.transformation()
                    m = QTransform()
                    m.translate(size.width() / 2, size.height() / 2)
                    if transf & QImageIOHandler.TransformationMirror:
                        # horizontal mirror
                        m.scale(-1, 1)
                    if transf & QImageIOHandler.TransformationFlip:
                        # vertical mirror
                        m.scale(1, -1)
                    if transf & QImageIOHandler.TransformationRotate90:
                        # rotate 90
                        m.rotate(-90)
                        m.translate(size.height() / -2, size.width() / -2)
                    else:
                        m.translate(size.width() / -2, size.height() / -2)
                    clip = m.mapRect(clip)
                reader.setClipRect(clip)
            return reader.read()


class ImagePage(page.AbstractRenderedPage):
    """A Page that displays an image in any file format supported by Qt."""
    autoTransform = True    # whether to automatically apply exif transformations
    dpi = 96   # TODO: maybe this can be image dependent.

    def __init__(self, container, renderer=None):
        super().__init__(renderer)
        self.setPageSize(container.size())
        self._ic = container

    @classmethod
    def load(cls, filename, renderer=None):
        """Load the image and yield one ImagePage instance if loading was successful."""
        loader = ImageLoader(filename, cls.autoTransform)
        if loader.size():
            yield cls(loader, renderer)

    @classmethod
    def fromImage(cls, image, renderer=None):
        """Instantiate one ImagePage from the supplied QImage.

        As the image is kept in memory, it is not advised to instantiate many
        Page instances this way. Use load() for images on the filesystem.
        The image must be valid, and have a size > 0.

        """
        return cls(ImageContainer(image), renderer)

    def print(self, painter, rect=None, paperColor=None):
        """Paint a page for printing."""
        if rect is None:
            image = self._ic.image()
        else:
            rect = rect.normalized() & self.pageRect()
            # we copy the image, because QSvgGenerator otherwise includes the
            # full image in the resulting SVG file!
            image = self._ic.image(rect.toRect())
        painter.drawImage(QPoint(0, 0), image)

    def image(self, rect=None, dpiX=None, dpiY=None, paperColor=None):
        """Return a QImage of the specified rectangle."""
        if rect is None:
            rect = self.rect()
        else:
            rect = rect & self.rect()
        if dpiX is None:
            dpiX = self.dpi
        if dpiY is None:
            dpiY = dpiX

        s = self.defaultSize()
        m = QTransform()
        m.scale(s.width() * dpiX / self.dpi, s.height() * dpiY / self.dpi)
        m.translate(.5, .5)
        m.rotate(self.computedRotation * 90)
        m.translate(-.5, -.5)
        m.scale(1 / self.pageWidth, 1 / self.pageHeight)

        source = self.transform().inverted()[0].mapRect(rect)
        return self._ic.image(source).transformed(m, Qt.SmoothTransformation)

    def group(self):
        return self._ic

    def mutex(self):
        return self._ic


class ImageDocument(document.MultiSourceDocument):
    """A Document representing a group of images.

    A source may be a filename, a QIODevice or a QImage.

    """
    pageClass = ImagePage

    def createPages(self):
        for s in self.sources():
            if isinstance(s, QImage):
                if not s.isNull():
                    yield self.pageClass.fromImage(s, self.renderer)
            else:
                for p in self.pageClass.load(s, self.renderer):
                    yield p


class ImageRenderer(render.AbstractRenderer):
    def draw(self, page, painter, key, tile, paperColor=None):
        """Draw the specified tile of the page (coordinates in key) on painter."""
        # determine the part to draw; convert tile to viewbox
        source = self.map(key, page.pageRect()).mapRect(QRect(*tile))
        target = QRect(0, 0, tile.w, tile.h)
        if key.rotation & 1:
            target.setSize(target.size().transposed())
        image = page._ic.image(source).scaled(
            target.size(), Qt.KeepAspectRatio, Qt.SmoothTransformation)
        painter.drawImage(target, image)



# install a default renderer, so SvgPage can be used directly
ImagePage.renderer = ImageRenderer()




# -*- coding: utf-8 -*-
#
# This file is part of the qpageview package.
#
# Copyright (c) 2016 - 2019 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
The View, deriving from QAbstractScrollArea.
"""

import collections
import contextlib
import weakref

from PyQt5.QtCore import pyqtSignal, QEvent, QPoint, QRect, QSize, Qt
from PyQt5.QtGui import QCursor, QPainter, QPalette, QRegion
from PyQt5.QtWidgets import QGestureEvent, QPinchGesture, QStyle
from PyQt5.QtPrintSupport import QPrinter, QPrintDialog

from . import layout
from . import page
from . import scrollarea
from . import util

from .constants import (

    # rotation:
    Rotate_0,
    Rotate_90,
    Rotate_180,
    Rotate_270,

    # viewModes:
    FixedScale,
    FitWidth,
    FitHeight,
    FitBoth,

    # orientation:
    Horizontal,
    Vertical,
)


Position = collections.namedtuple("Position", "pageNumber x y")


class View(scrollarea.ScrollArea):
    """View is a generic scrollable widget to display Pages in a layout.

    Using setPageLayout() you can set a PageLayout to the View, and you can
    add Pages to the layout using a list-like api. (PageLayout derives from
    list). A simple PageLayout is set by default. Call updatePageLayout() after
    every change to the layout (like adding or removing pages).

    You can also add a Magnifier to magnify parts of a Page, and a Rubberband
    to enable selecting a rectangular region.

    View emits the following signals:

    :attr:`pageCountChanged` (int)
        emitted when the total amount of pages has changed

    :attr:`currentPageNumberChanged` (int)
        emitted when the current page number has changed (starting with 1)

    :attr:`viewModeChanged` (int)
        emitted when the ``viewMode`` has changed

    :attr:`rotationChanged` (int)
        emitted when the ``rotation`` has changed

    :attr:`orientationChanged` (int)
        emitted when the ``orientation`` has changed

    :attr:`zoomFactorChanged` (float)
        emitted when the ``zoomFactor`` has changed

    :attr:`continuousModeChanged` (bool)
        emitted when the ``continuousMode`` has changed

    :attr:`pageLayoutModeChanged` (str)
        emitted when the ``pageLayoutMode`` has changed

    :attr:`pageLayoutUpdated` ()
        emitted whenever the page layout has been updated (redraw/resize)

    """

    MIN_ZOOM = 0.05
    MAX_ZOOM = 64.0

    #: whether to enable mouse wheel zooming
    wheelZoomingEnabled = True

    #: whether to enable kinetic scrolling while paging (setCurrentPageNumber)
    kineticPagingEnabled = True

    #: whether to keep track of current page while scrolling
    pagingOnScrollEnabled = True

    #: whether a mouse click in a page makes it the current page
    clickToSetCurrentPageEnabled = True

    #: whether PageUp and PageDown call setCurrentPageNumber instead of scroll
    strictPagingEnabled = False

    #: can be set to a DocumentPropertyStore object. If set, the object is
    #: used to store certain View settings on a per-document basis.
    #: (This happens in the :meth:`clear` and :meth:`setDocument` methods.)
    documentPropertyStore = None

    #: (int) emitted when the total amount of pages has changed
    pageCountChanged = pyqtSignal(int)

    #: (int) emitted when the current page number has changed (starting with 1)
    currentPageNumberChanged = pyqtSignal(int)

    #: (int) emitted when the ``viewMode`` has changed
    viewModeChanged = pyqtSignal(int)

    #: (int) emitted when the ``rotation`` has changed
    rotationChanged = pyqtSignal(int)

    #: (int) emitted when the ``orientation`` has changed
    orientationChanged = pyqtSignal(int)

    #: (float) emitted when the ``zoomFactor`` has changed
    zoomFactorChanged = pyqtSignal(float)

    #: (bool) emitted when the ``continuousMode`` has changed
    continuousModeChanged = pyqtSignal(bool)

    #: (str) emitted when the ``pageLayoutMode`` has changed
    pageLayoutModeChanged = pyqtSignal(str)

    #: emitted whenever the page layout has been updated (redraw/resize)
    pageLayoutUpdated = pyqtSignal()

    def __init__(self, parent=None, **kwds):
        super().__init__(parent, **kwds)
        self._document = None
        self._currentPageNumber = 0
        self._pageCount = 0
        self._scrollingToPage = 0
        self._prev_pages_to_paint = set()
        self._viewMode = FixedScale
        self._pageLayout = None
        self._magnifier = None
        self._rubberband = None
        self._pinchStartFactor = None
        self.grabGesture(Qt.PinchGesture)
        self.viewport().setBackgroundRole(QPalette.Dark)
        self.verticalScrollBar().setSingleStep(20)
        self.horizontalScrollBar().setSingleStep(20)
        self.setMouseTracking(True)
        self.setMinimumSize(QSize(60, 60))
        self.setPageLayout(layout.PageLayout())
        props = self.properties().setdefaults()
        self._viewMode = props.viewMode
        self._pageLayout.continuousMode = props.continuousMode
        self._pageLayout.orientation = props.orientation
        self._pageLayoutMode = props.pageLayoutMode
        self.pageLayout().engine = self.pageLayoutModes()[props.pageLayoutMode]()

    def pageCount(self):
        """Return the number of pages in the view."""
        return self._pageCount

    def currentPageNumber(self):
        """Return the current page number in view (starting with 1)."""
        return self._currentPageNumber

    def setCurrentPageNumber(self, num):
        """Scrolls to the specified page number (starting with 1).

        If the page is already in view, the view is not scrolled, otherwise
        the view is scrolled to center the page. (If the page is larger than
        the view, the top-left corner is positioned top-left in the view.)

        """
        self.updateCurrentPageNumber(num)
        page = self.currentPage()
        if page:
            margins = self._pageLayout.margins() + self._pageLayout.pageMargins()
            with self.pagingOnScrollDisabled():
                self.ensureVisible(page.geometry(), margins, self.kineticPagingEnabled)
            if self.isScrolling():
                self._scrollingToPage = True

    def updateCurrentPageNumber(self, num):
        """Set the current page number without scrolling the view."""
        count = self.pageCount()
        n = max(min(count, num), 1 if count else 0)
        if n == num and n != self._currentPageNumber:
            self._currentPageNumber = num
            self.currentPageNumberChanged.emit(num)

    def gotoNextPage(self):
        """Convenience method to go to the next page."""
        num = self.currentPageNumber()
        if num < self.pageCount():
            self.setCurrentPageNumber(num + 1)

    def gotoPreviousPage(self):
        """Convenience method to go to the previous page."""
        num = self.currentPageNumber()
        if num > 1:
            self.setCurrentPageNumber(num - 1)

    def currentPage(self):
        """Return the page pointed to by currentPageNumber()."""
        if self._pageCount:
            return self._pageLayout[self._currentPageNumber-1]

    def page(self, num):
        """Return the page at the specified number (starting at 1)."""
        if 0 < num <= self._pageCount:
            return self._pageLayout[num-1]

    def pages(self):
        """Return a list of all Pages in the page layout."""
        return list(self._pageLayout)

    def position(self):
        """Return a three-tuple Position(pageNumber, x, y).

        The Position describes where the center of the viewport is on the layout.
        The page is the page number (starting with 1) and x and y the position
        on the page, in a 0..1 range. This way a position can be remembered even
        if the zoom or orientation of the layout changes.

        """
        pos = self.viewport().rect().center()
        i, x, y = self._pageLayout.pos2offset(pos - self.layoutPosition())
        return Position(i + 1, x, y)

    def setPosition(self, position, allowKinetic=True):
        """Centers the view on the spot stored in the specified Position.

        If allowKinetic is False, immediately jumps to the position, otherwise
        scrolls smoothly (if kinetic scrolling is enabled).

        """
        i, x, y = position
        rect = self.viewport().rect()
        rect.moveCenter(self._pageLayout.offset2pos((i - 1, x, y)))
        self.ensureVisible(rect, allowKinetic=allowKinetic)

    def setPageLayout(self, layout):
        """Set our current PageLayout instance.

        The dpiX and dpiY attributes of the layout are set to the physical
        resolution of the widget, which should result in a natural size of 100%
        at zoom factor 1.0.

        """
        if self._pageLayout:
            self._unschedulePages(self._pageLayout)
        layout.dpiX = self.physicalDpiX()
        layout.dpiY = self.physicalDpiY()
        self._pageLayout = layout
        self.updatePageLayout()

    def pageLayout(self):
        """Return our current PageLayout instance."""
        return self._pageLayout

    def pageLayoutModes(self):
        """Return a dictionary mapping names to callables.

        The callable returns a configured LayoutEngine that is set to the
        page layout. You can reimplement this method to returns more layout
        modes, but it is required that the name "single" exists.

        """
        def single():
            return layout.LayoutEngine()

        def raster():
            return layout.RasterLayoutEngine()

        def double_left():
            engine = layout.RowLayoutEngine()
            engine.pagesPerRow = 2
            engine.pagesFirstRow = 0
            return engine

        def double_right():
            engine = double_left()
            engine.pagesFirstRow = 1
            return engine

        return locals()

    def pageLayoutMode(self):
        """Return the currently set page layout mode."""
        return self._pageLayoutMode

    def setPageLayoutMode(self, mode):
        """Set the page layout mode.

        The mode is one of the names returned by pageLayoutModes().
        The mode name "single" is guaranteed to exist.

        """
        if mode != self._pageLayoutMode:
            # get a suitable LayoutEngine
            try:
                engine = self.pageLayoutModes()[mode]()
            except KeyError:
                return
            self._pageLayout.engine = engine
            # keep the current page in view
            page = self.currentPage()
            self.updatePageLayout()
            if page:
                margins = self._pageLayout.margins() + self._pageLayout.pageMargins()
                with self.pagingOnScrollDisabled():
                    self.ensureVisible(page.geometry(), margins, False)
            self._pageLayoutMode = mode
            self.pageLayoutModeChanged.emit(mode)
            if self.viewMode():
                with self.keepCentered():
                    self.fitPageLayout()

    def updatePageLayout(self, lazy=False):
        """Update layout, adjust scrollbars, keep track of page count.

        If lazy is set to True, calls lazyUpdate() to update the view.

        """
        self._pageLayout.update()

        # keep track of page count
        count = self._pageLayout.count()
        if count != self._pageCount:
            self._pageCount = count
            self.pageCountChanged.emit(count)
            n = max(min(count, self._currentPageNumber), 1 if count else 0)
            self.updateCurrentPageNumber(n)

        self.setAreaSize(self._pageLayout.size())
        self.pageLayoutUpdated.emit()
        self.lazyUpdate() if lazy else self.viewport().update()

    @contextlib.contextmanager
    def modifyPages(self):
        """Return the list of pages and enter a context to make modifications.

        Note that the first page is at index 0.
        On exit of the context the page layout is updated.

        """
        pages = list(self._pageLayout)
        if self.rubberband():
            selectedpages = set(p for p, r in self.rubberband().selectedPages())
        else:
            selectedpages = set()
        lazy = bool(pages)
        try:
            yield pages
        finally:
            lazy &= bool(pages)
            removedpages = set(self._pageLayout) - set(pages)
            if selectedpages & removedpages:
                self.rubberband().clearSelection() # rubberband'll always be there
            self._unschedulePages(removedpages)
            self._pageLayout[:] = pages
            if self._viewMode:
                zoomFactor = self._pageLayout.zoomFactor
                self.fitPageLayout()
                if zoomFactor != self._pageLayout.zoomFactor:
                    lazy = False
            self.updatePageLayout(lazy)

    @contextlib.contextmanager
    def modifyPage(self, num):
        """Return the page (numbers start with 1) and enter a context.

        On exit of the context, the page layout is updated.

        """
        page = self.page(num)
        yield page
        if page:
            self._unschedulePages((page,))
            self.updatePageLayout(True)

    def clear(self):
        """Convenience method to clear the current layout."""
        self.setPages([])

    def setPages(self, pages):
        """Load the iterable of pages into the View.

        Existing pages are removed, and the document is set to None.

        """
        if self.documentPropertyStore and self._document:
            self.documentPropertyStore.set(self._document, self.properties().get(self))
        self._document = None
        with self.modifyPages() as pgs:
            pgs[:] = pages

    def setDocument(self, document):
        """Set the Document to display (see document.Document)."""
        store = self._document is not document and self.documentPropertyStore
        if store and self._document:
            store.set(self._document, self.properties().get(self))
        self._document = document
        with self.modifyPages() as pages:
            pages[:] = document.pages()
        if store:
            (store.get(document) or store.default or self.properties()).set(self)

    def document(self):
        """Return the Document currently displayed (see document.Document)."""
        return self._document

    def reload(self):
        """If a Document was set, invalidate()s it and then reloads it."""
        if self._document:
            self._document.invalidate()
            with self.modifyPages() as pages:
                pages[:] = self._document.pages()

    def loadPdf(self, filename, renderer=None):
        """Convenience method to load the specified PDF file.

        The filename can also be a QByteArray or an already loaded
        popplerqt5.Poppler.Document instance.

        """
        from . import poppler
        self.setDocument(poppler.PopplerDocument(filename, renderer))

    def loadSvgs(self, filenames, renderer=None):
        """Convenience method to load the specified list of SVG files.

        Each SVG file is loaded in one Page. A filename can also be a
        QByteArray.

        """
        from . import svg
        self.setDocument(svg.SvgDocument(filenames, renderer))

    def loadImages(self, filenames, renderer=None):
        """Convenience method to load images from the specified list of files.

        Each image is loaded in one Page. A filename can also be a
        QByteArray or a QImage.

        """
        from . import image
        self.setDocument(image.ImageDocument(filenames, renderer))

    def print(self, printer=None, pageNumbers=None, showDialog=True):
        """Print all, or speficied pages to QPrinter printer.

        If given the pageNumbers should be a list containing page numbers
        starting with 1. If showDialog is True, a print dialog is shown, and
        printing is canceled when the user cancels the dialog.

        If the QPrinter to use is not specified, a default one is created.
        The print job is started and returned (a printing.PrintJob instance),
        so signals for monitoring the progress could be connected to. (If the
        user cancels the dialog, no print job is returned.)

        """
        if printer is None:
            printer = QPrinter()
            printer.setResolution(300)
        if showDialog:
            dlg = QPrintDialog(printer, self)
            dlg.setMinMax(1, self.pageCount())
            if not dlg.exec_():
                return  # cancelled
        if not pageNumbers:
            if printer.printRange() == QPrinter.CurrentPage:
                pageNumbers = [self.currentPageNumber()]
            else:
                if printer.printRange() == QPrinter.PageRange:
                    first = printer.toPage() or 1
                    last = printer.fromPage() or self.pageCount()
                else:
                    first, last = 1, self.pageCount()
                pageNumbers = list(range(first, last + 1))
            if printer.pageOrder() == QPrinter.LastPageFirst:
                pageNumbers.reverse()
        # add the page objects
        pageList = [(n, self.page(n)) for n in pageNumbers]
        from . import printing
        job = printing.PrintJob(printer, pageList)
        job.start()
        return job

    @staticmethod
    def properties():
        """Return an uninitialized ViewProperties object."""
        return ViewProperties()

    def readProperties(self, settings):
        """Read View settings from the QSettings object.

        If a documentPropertyStore is set, the settings are also set
        as default for the DocumentPropertyStore.

        """
        props = self.properties().load(settings)
        props.position = None   # storing the position makes no sense
        props.set(self)
        if self.documentPropertyStore:
            self.documentPropertyStore.default = props

    def writeProperties(self, settings):
        """Write the current View settings to the QSettings object.

        If a documentPropertyStore is set, the settings are also set
        as default for the DocumentPropertyStore.

        """
        props = self.properties().get(self)
        props.position = None   # storing the position makes no sense
        props.save(settings)
        if self.documentPropertyStore:
            self.documentPropertyStore.default = props

    def setViewMode(self, mode):
        """Sets the current ViewMode."""
        if mode == self._viewMode:
            return
        self._viewMode = mode
        if mode:
            with self.keepCentered():
                self.fitPageLayout()
        else:
            # call layout once to tell FixedScale is active
            self.pageLayout().fit(QSize(), mode)
        self.viewModeChanged.emit(mode)

    def viewMode(self):
        """Returns the current ViewMode."""
        return self._viewMode

    def setRotation(self, rotation):
        """Set the current rotation."""
        layout = self._pageLayout
        if rotation != layout.rotation:
            with self.keepCentered():
                layout.rotation = rotation
                self.fitPageLayout()
            self.rotationChanged.emit(rotation)

    def rotation(self):
        """Return the current rotation."""
        return self._pageLayout.rotation

    def rotateLeft(self):
        """Rotate the pages 270 degrees."""
        self.setRotation((self.rotation() - 1) & 3)

    def rotateRight(self):
        """Rotate the pages 90 degrees."""
        self.setRotation((self.rotation() + 1) & 3)

    def setOrientation(self, orientation):
        """Set the orientation (Horizontal or Vertical)."""
        layout = self._pageLayout
        if orientation != layout.orientation:
            with self.keepCentered():
                layout.orientation = orientation
                self.fitPageLayout()
            self.orientationChanged.emit(orientation)

    def orientation(self):
        """Return the current orientation (Horizontal or Vertical)."""
        return self._pageLayout.orientation

    def setContinuousMode(self, continuous):
        """Sets whether the layout should display all pages.

        If True, the layout shows all pages. If False, only the page set
        containing the current page is displayed. If the pageLayout() does not
        support the PageSetLayoutMixin methods, this method does nothing.

        """
        layout = self._pageLayout
        oldcontinuous = layout.continuousMode
        if continuous:
            if not oldcontinuous:
                with self.pagingOnScrollDisabled(), self.keepCentered():
                    layout.continuousMode = True
                    self.fitPageLayout()
                self.continuousModeChanged.emit(True)
        elif oldcontinuous:
            p = self.currentPage()
            index = layout.index(p) if p else 0
            with self.pagingOnScrollDisabled(), self.keepCentered():
                layout.continuousMode = False
                layout.currentPageSet = layout.pageSet(index)
                self.fitPageLayout()
            self.continuousModeChanged.emit(False)

    def continuousMode(self):
        """Return True if the layout displays all pages."""
        return self._pageLayout.continuousMode

    def displayPageSet(self, what):
        """Try to display a page set (if the layout is not in continuous mode).

        `what` can be:

            "next":     go to the next page set
            "previous": go to the previous page set
            "first":    go to the first page set
            "last":     go to the last page set
            integer:    go to the specified page set

        """
        layout = self._pageLayout
        if layout.continuousMode:
            return

        sb = None  # where to move the scrollbar after fitlayout
        if what == "first":
            what = 0
            sb = "up"   # move to the start
        elif what == "last":
            what = layout.pageSetCount() - 1
            sb = "down" # move to the end
        elif what == "previous":
            what = layout.currentPageSet - 1
            if what < 0:
                return
            sb = "down"
        elif what == "next":
            what = layout.currentPageSet + 1
            if what >= layout.pageSetCount():
                return
            sb = "up"
        elif not 0 <= what < layout.pageSetCount():
            return
        layout.currentPageSet = what
        self.fitPageLayout()
        self.updatePageLayout()
        if sb:
            self.verticalScrollBar().setValue(0 if sb == "up" else self.verticalScrollBar().maximum())
        if self.pagingOnScrollEnabled and not self._scrollingToPage:
            s = layout.currentPageSetSlice()
            num = s.stop - 1 if sb == "down" else s.start
            self.updateCurrentPageNumber(num + 1)

    def setMagnifier(self, magnifier):
        """Sets the Magnifier to use (or None to disable the magnifier).

        The viewport takes ownership of the Magnifier.

        """
        if self._magnifier:
            self.viewport().removeEventFilter(self._magnifier)
            self._magnifier.setParent(None)
        self._magnifier = magnifier
        if magnifier:
            magnifier.setParent(self.viewport())
            self.viewport().installEventFilter(magnifier)

    def magnifier(self):
        """Returns the currently set magnifier."""
        return self._magnifier

    def setRubberband(self, rubberband):
        """Sets the Rubberband to use for selections (or None to not use one)."""
        if self._rubberband:
            self.viewport().removeEventFilter(self._rubberband)
            self.zoomFactorChanged.disconnect(self._rubberband.slotZoomChanged)
            self.rotationChanged.disconnect(self._rubberband.clearSelection)
            self._rubberband.setParent(None)
        self._rubberband = rubberband
        if rubberband:
            rubberband.setParent(self.viewport())
            rubberband.clearSelection()
            self.viewport().installEventFilter(rubberband)
            self.zoomFactorChanged.connect(rubberband.slotZoomChanged)
            self.rotationChanged.connect(rubberband.clearSelection)

    def rubberband(self):
        """Return the currently set rubberband."""
        return self._rubberband

    @contextlib.contextmanager
    def pagingOnScrollDisabled(self):
        """During this context a scroll is not tracked to update the current page number."""
        old, self._scrollingToPage = self._scrollingToPage, True
        try:
            yield
        finally:
            self._scrollingToPage = old

    def scrollContentsBy(self, dx, dy):
        """Reimplemented to move the rubberband and adjust the mouse cursor."""
        if self._rubberband:
            self._rubberband.scrollBy(QPoint(dx, dy))
        if not self.isScrolling() and not self.isDragging():
            # don't adjust the cursor during a kinetic scroll
            pos = self.viewport().mapFromGlobal(QCursor.pos())
            if pos in self.viewport().rect() and not self.viewport().childAt(pos):
                self.adjustCursor(pos)
        self.viewport().update()

        # keep track of current page. If the scroll wasn't initiated by the
        # setCurrentPage() call, check # whether the current page number needs
        # to be updated
        if self.pagingOnScrollEnabled and not self._scrollingToPage and self.pageCount() > 0:
            # do nothing if current page is still fully in view
            if self.currentPage().geometry() not in self.visibleRect():
                # find the page in the center of the view
                layout = self._pageLayout
                pos = self.visibleRect().center()
                p = layout.pageAt(pos) or layout.nearestPageAt(pos)
                if p:
                    num = layout.index(p) + 1
                    self.updateCurrentPageNumber(num)

    def stopScrolling(self):
        """Reimplemented to adjust the mouse cursor on scroll stop."""
        super().stopScrolling()
        self._scrollingToPage = False
        pos = self.viewport().mapFromGlobal(QCursor.pos())
        if pos in self.viewport().rect() and not self.viewport().childAt(pos):
            self.adjustCursor(pos)

    def fitPageLayout(self):
        """Fit the layout according to the view mode.

        Does nothing in FixedScale mode. Prevents scrollbar/resize loops by
        precalculating which scrollbars will appear.

        """
        mode = self.viewMode()
        if mode == FixedScale:
            return

        maxsize = self.maximumViewportSize()

        # can vertical or horizontal scrollbars appear?
        vcan = self.verticalScrollBarPolicy() == Qt.ScrollBarAsNeeded
        hcan = self.horizontalScrollBarPolicy() == Qt.ScrollBarAsNeeded

        # width a scrollbar takes off the viewport size
        framewidth = 0
        if self.style().styleHint(QStyle.SH_ScrollView_FrameOnlyAroundContents, None, self):
            framewidth = self.style().pixelMetric(QStyle.PM_DefaultFrameWidth) * 2
        scrollbarextent = self.style().pixelMetric(QStyle.PM_ScrollBarExtent, None, self) + framewidth

        # remember old factor
        zoom_factor = self.zoomFactor()

        # first try to fit full size
        layout = self._pageLayout
        layout.fit(maxsize, mode)
        layout.update()

        # minimal values
        minwidth = maxsize.width()
        minheight = maxsize.height()
        if vcan:
            minwidth -= scrollbarextent
        if hcan:
            minheight -= scrollbarextent

        # do width and/or height fit?
        fitw = layout.width <= maxsize.width()
        fith = layout.height <= maxsize.height()

        if not fitw and not fith:
            if vcan or hcan:
                layout.fit(QSize(minwidth, minheight), mode)
        elif mode & FitWidth and fitw and not fith and vcan:
            # a vertical scrollbar will appear
            w = minwidth
            layout.fit(QSize(w, maxsize.height()), mode)
            layout.update()
            if layout.height <= maxsize.height():
                # now the vert. scrollbar would disappear!
                # enlarge it as long as the vertical scrollbar would not be needed
                while True:
                    w += 1
                    layout.fit(QSize(w, maxsize.height()), mode)
                    layout.update()
                    if layout.height > maxsize.height():
                        layout.fit(QSize(w - 1, maxsize.height()), mode)
                        break
        elif mode & FitHeight and fith and not fitw and hcan:
            # a horizontal scrollbar will appear
            h = minheight
            layout.fit(QSize(maxsize.width(), h), mode)
            layout.update()
            if layout.width <= maxsize.width():
                # now the horizontal scrollbar would disappear!
                # enlarge it as long as the horizontal scrollbar would not be needed
                while True:
                    h += 1
                    layout.fit(QSize(maxsize.width(), h), mode)
                    layout.update()
                    if layout.width > maxsize.width():
                        layout.fit(QSize(maxsize.width(), h - 1), mode)
                        break
        if zoom_factor != self.zoomFactor():
            self.zoomFactorChanged.emit(self.zoomFactor())
            self._unschedulePages(layout)

    @contextlib.contextmanager
    def keepCentered(self, pos=None):
        """Context manager to keep the same spot centered while changing the layout.

        If pos is not given, the viewport's center is used.
        After yielding, updatePageLayout() is called.

        """
        if pos is None:
            pos = self.viewport().rect().center()

        # find the spot on the page
        layout = self._pageLayout
        layout_pos = self.layoutPosition()
        pos_on_layout = pos - layout_pos
        offset = layout.pos2offset(pos_on_layout)
        pos_on_layout -= layout.pos()   # pos() of the layout might change

        yield
        self.updatePageLayout()

        new_pos_on_layout = layout.offset2pos(offset) - layout.pos()
        diff = new_pos_on_layout - pos
        self.verticalScrollBar().setValue(diff.y())
        self.horizontalScrollBar().setValue(diff.x())

    def setZoomFactor(self, factor, pos=None):
        """Set the zoom factor (1.0 by default).

        If pos is given, that position (in viewport coordinates) is kept in the
        center if possible. If None, zooming centers around the viewport center.

        """
        factor = max(self.MIN_ZOOM, min(self.MAX_ZOOM, factor))
        if factor != self._pageLayout.zoomFactor:
            with self.keepCentered(pos):
                self._pageLayout.zoomFactor = factor
            if self._pageLayout.zoomsToFit():
                self.setViewMode(FixedScale)
            self.zoomFactorChanged.emit(factor)
            self._unschedulePages(self._pageLayout)

    def zoomFactor(self):
        """Return the page layout's zoom factor."""
        return self._pageLayout.zoomFactor

    def zoomIn(self, pos=None, factor=1.1):
        """Zoom in.

        If pos is given, it is the position in the viewport to keep centered.
        Otherwise zooming centers around the viewport center.

        """
        self.setZoomFactor(self.zoomFactor() * factor, pos)

    def zoomOut(self, pos=None, factor=1.1):
        """Zoom out.

        If pos is given, it is the position in the viewport to keep centered.
        Otherwise zooming centers around the viewport center.

        """
        self.setZoomFactor(self.zoomFactor() / factor, pos)

    def zoomNaturalSize(self, pos=None):
        """Zoom to the natural pixel size of the current page.

        The natural pixel size zoom factor can be different than 1.0, if the
        screen's DPI differs from the current page's DPI.

        """
        p = self.currentPage()
        factor = p.dpi / self.physicalDpiX() if p else 1.0
        self.setZoomFactor(factor, pos)

    def layoutPosition(self):
        """Return the position of the PageLayout relative to the viewport.

        This is the top-left position of the layout, relative to the
        top-left position of the viewport.

        If the layout is smaller than the viewport it is centered by default.
        (See ScrollArea.alignment.)

        """
        return self.areaPos() - self._pageLayout.pos()

    def visibleRect(self):
        """Return the QRect of the page layout that is currently visible in the viewport."""
        return self.visibleArea().translated(self._pageLayout.pos())

    def visiblePages(self, rect=None):
        """Yield the Page instances that are currently visible.

        If rect is not given, the visibleRect() is used.  The pages are sorted
        so that the pages with the largest visible part come first.

        """
        if rect is None:
            rect = self.visibleRect()
        def key(page):
            overlayrect = rect & page.geometry()
            return overlayrect.width() * overlayrect.height()
        return sorted(self._pageLayout.pagesAt(rect), key=key, reverse=True)

    def ensureVisible(self, rect, margins=None, allowKinetic=True):
        """Ensure rect is visible, switching page set if necessary."""
        if not any(self.pageLayout().pagesAt(rect)):
            if self.continuousMode():
                return
            # we might need to switch page set
            # find the rect
            for p in layout.PageRects(self.pageLayout()).intersecting(*rect.getCoords()):
                num = self.pageLayout().index(p)
                self.displayPageSet(self.pageLayout().pageSet(num))
                break
            else:
                return
        rect = rect.translated(-self._pageLayout.pos())
        super().ensureVisible(rect, margins, allowKinetic)

    def adjustCursor(self, pos):
        """Sets the correct mouse cursor for the position on the page."""
        pass

    def repaintPage(self, page):
        """Call this when you want to redraw the specified page."""
        rect = page.geometry().translated(self.layoutPosition())
        self.viewport().update(rect)

    def lazyUpdate(self, page=None):
        """Lazily repaint page (if visible) or all visible pages.

        Defers updating the viewport for a page until all rendering tasks for
        that page have finished. This reduces flicker.

        """
        viewport = self.viewport()
        full = True
        updates = []
        for p in self.visiblePages():
            rect = self.visibleRect() & p.geometry()
            if rect and p.renderer:
                info = p.renderer.info(p, viewport, rect.translated(-p.pos()))
                if info.missing:
                    full = False
                    if page is p or page is None:
                        p.renderer.schedule(p, info.key, info.missing, self.lazyUpdate)
                elif page is p or page is None:
                    updates.append(rect.translated(self.layoutPosition()))
        if full:
            viewport.update()
        elif updates:
            viewport.update(sum(updates, QRegion()))

    def rerender(self, page=None):
        """Schedule the specified page or all pages for rerendering.

        Call this when you have changed render options or page contents.
        Repaints the page or visible pages lazily, reducing flicker.

        """
        renderers = collections.defaultdict(list)
        pages = (page,) if page else self._pageLayout
        for p in pages:
            if p.renderer:
                renderers[p.renderer].append(p)
        for renderer, pages in renderers.items():
            renderer.invalidate(pages)
        self.lazyUpdate(page)

    def _unschedulePages(self, pages):
        """(Internal.)
        Unschedule rendering of pages that are pending but not needed anymore.

        Called inside paintEvent, on zoomFactor change and some other places.
        This prevents rendering jobs hogging the cpu for pages that are deleted
        or out of view.

        """
        unschedule = collections.defaultdict(set)
        for page in pages:
            if page.renderer:
                unschedule[page.renderer].add(page)
        for renderer, pages in unschedule.items():
            renderer.unschedule(pages, self.repaintPage)

    def pagesToPaint(self, rect, painter):
        """Yield (page, rect) to paint in the specified rectangle.

        The specified rect is in viewport coordinates, as in the paint event.
        The returned rect describes the part of the page actually to draw, in
        page coordinates. (The full rect can be found in page.rect().)
        Translates the painter to the top left of each page.

        The pages are sorted with largest area last.

        """
        layout_pos = self.layoutPosition()
        ev_rect = rect.translated(-layout_pos)
        for p in self.visiblePages(ev_rect):
            r = (p.geometry() & ev_rect).translated(-p.pos())
            painter.save()
            painter.translate(layout_pos + p.pos())
            yield p, r
            painter.restore()

    def event(self, ev):
        """Reimplemented to get Gesture events."""
        if isinstance(ev, QGestureEvent) and self.handleGestureEvent(ev):
            ev.accept() # Accepts all gestures in the event
            return True
        return super().event(ev)

    def handleGestureEvent(self, event):
        """Gesture event handler.

        Return False if event is not accepted. Currently only cares about
        PinchGesture. Could also handle Swipe and Pan gestures.

        """
        ## originally contributed by David Rydh, 2017
        pinch = event.gesture(Qt.PinchGesture)
        if pinch:
            return self.pinchGesture(pinch)
        return False

    def pinchGesture(self, gesture):
        """Pinch gesture event handler.

        Return False if event is not accepted. Currently only cares about
        ScaleFactorChanged and not RotationAngleChanged.

        """
        ## originally contributed by David Rydh, 2017
        # Gesture start? Reset _pinchStartFactor in case we didn't
        # catch the finish event
        if gesture.state() == Qt.GestureStarted:
            self._pinchStartFactor = None

        changeFlags = gesture.changeFlags()
        if changeFlags & QPinchGesture.ScaleFactorChanged:
            factor = gesture.property("totalScaleFactor")
            if not self._pinchStartFactor: # Gesture start?
                self._pinchStartFactor = self.zoomFactor()
            self.setZoomFactor(self._pinchStartFactor * factor,
                      self.mapFromGlobal(gesture.hotSpot().toPoint()))

        # Gesture finished?
        if gesture.state() in (Qt.GestureFinished, Qt.GestureCanceled):
            self._pinchStartFactor = None

        return True

    def paintEvent(self, ev):
        """Paint the contents of the viewport."""
        painter = QPainter(self.viewport())
        pages_to_paint = set()
        for p, r in self.pagesToPaint(ev.rect(), painter):
            p.paint(painter, r, self.repaintPage)
            pages_to_paint.add(p)

        # remove pending render jobs for pages that were visible, but are not
        # visible now
        rect = self.visibleRect()
        pages = set(page
            for page in self._prev_pages_to_paint - pages_to_paint
                if not rect.intersects(page.geometry()))
        self._unschedulePages(pages)
        self._prev_pages_to_paint = pages_to_paint

    def resizeEvent(self, ev):
        """Reimplemented to scale the view if needed and update the scrollbars."""
        if self._viewMode and not self._pageLayout.empty():
            with self.pagingOnScrollDisabled():
                # sensible repositioning
                vbar = self.verticalScrollBar()
                hbar = self.horizontalScrollBar()
                x, xm = hbar.value(), hbar.maximum()
                y, ym = vbar.value(), vbar.maximum()
                self.fitPageLayout()
                self.updatePageLayout()
                if xm: hbar.setValue(round(x * hbar.maximum() / xm))
                if ym: vbar.setValue(round(y * vbar.maximum() / ym))
        super().resizeEvent(ev)

    def wheelEvent(self, ev):
        """Reimplemented to support wheel zooming and paging through page sets."""
        if self.wheelZoomingEnabled and ev.angleDelta().y() and ev.modifiers() & Qt.CTRL:
            factor = 1.1 ** util.sign(ev.angleDelta().y())
            self.setZoomFactor(self.zoomFactor() * factor, ev.pos())
        elif not ev.modifiers():
            # if scrolling is not possible, try going to next or previous pageset.
            sb = self.verticalScrollBar()
            sp = self.strictPagingEnabled
            if ev.angleDelta().y() > 0 and sb.value() == 0:
                self.gotoPreviousPage() if sp else self.displayPageSet("previous")
            elif ev.angleDelta().y() < 0 and sb.value() == sb.maximum():
                self.gotoNextPage() if sp else self.displayPageSet("next")
            else:
                super().wheelEvent(ev)
        else:
            super().wheelEvent(ev)

    def mousePressEvent(self, ev):
        """Implemented to set the clicked page as current, without moving it."""
        if self.clickToSetCurrentPageEnabled:
            page = self._pageLayout.pageAt(ev.pos() - self.layoutPosition())
            if page:
                num = self._pageLayout.index(page) + 1
                self.updateCurrentPageNumber(num)
        super().mousePressEvent(ev)

    def mouseMoveEvent(self, ev):
        """Implemented to adjust the mouse cursor depending on the page contents."""
        # no cursor updates when dragging the background is busy, see scrollarea.py.
        if not self.isDragging():
            self.adjustCursor(ev.pos())
        super().mouseMoveEvent(ev)

    def keyPressEvent(self, ev):
        """Reimplemented to go to next or previous page set if possible."""
        # ESC clears the selection, if any.
        if (ev.key() == Qt.Key_Escape and not ev.modifiers()
            and self.rubberband() and self.rubberband().hasSelection()):
            self.rubberband().clearSelection()
            return

        # Paging through page sets?
        sb = self.verticalScrollBar()
        sp = self.strictPagingEnabled
        if ev.key() == Qt.Key_PageUp:
            if sp:
                self.gotoPreviousPage()
            elif sb.value() == 0:
                self.displayPageSet("previous")
            else:
                super().keyPressEvent(ev)
        elif ev.key() == Qt.Key_PageDown:
            if sp:
                self.gotoNextPage()
            elif sb.value() == sb.maximum():
                self.displayPageSet("next")
            else:
                super().keyPressEvent(ev)
        elif ev.key() == Qt.Key_Home and ev.modifiers() == Qt.ControlModifier:
            self.setCurrentPageNumber(1) if sp else self.displayPageSet("first")
        elif ev.key() == Qt.Key_End and ev.modifiers() == Qt.ControlModifier:
            self.setCurrentPageNumber(self.pageCount()) if sp else self.displayPageSet("last")
        else:
            super().keyPressEvent(ev)


class ViewProperties:
    """Simple helper class encapsulating certain settings of a View.

    The settings can be set to and got from a View, and saved to or loaded
    from a QSettings group.

    Class attributes serve as default values, None means: no change.
    All methods return self, so operations can easily be chained.

    If you inherit from a View and add more settings, you can also add
    properties to this class by inheriting from it. Reimplement
    View.properties() to return an instance of your new ViewProperties
    subclass.

    """
    position = None
    rotation = Rotate_0
    zoomFactor = 1.0
    viewMode = FixedScale
    orientation = None
    continuousMode = None
    pageLayoutMode = None

    def setdefaults(self):
        """Set all properties to default values. Also used by View on init."""
        self.orientation = Vertical
        self.continuousMode = True
        self.pageLayoutMode = "single"
        return self

    def copy(self):
        """Return a copy or ourselves."""
        cls = type(self)
        props = cls.__new__(cls)
        props.__dict__.update(self.__dict__)
        return props

    def names(self):
        """Return a tuple with all the property names we support."""
        return (
            'position',
            'rotation',
            'zoomFactor',
            'viewMode',
            'orientation',
            'continuousMode',
            'pageLayoutMode',
        )

    def mask(self, names):
        """Set properties not listed in names to None."""
        for name in self.names():
            if name not in names and getattr(self, name) is not None:
                setattr(self, name, None)
        return self

    def get(self, view):
        """Get the properties of a View."""
        self.position = view.position()
        self.rotation = view.rotation()
        self.orientation = view.orientation()
        self.viewMode = view.viewMode()
        self.zoomFactor = view.zoomFactor()
        self.continuousMode = view.continuousMode()
        self.pageLayoutMode = view.pageLayoutMode()
        return self

    def set(self, view):
        """Set all our properties that are not None to a View."""
        if self.pageLayoutMode is not None:
            view.setPageLayoutMode(self.pageLayoutMode)
        if self.rotation is not None:
            view.setRotation(self.rotation)
        if self.orientation is not None:
            view.setOrientation(self.orientation)
        if self.continuousMode is not None:
            view.setContinuousMode(self.continuousMode)
        if self.viewMode is not None:
            view.setViewMode(self.viewMode)
        if self.zoomFactor is not None:
            if self.viewMode is FixedScale or not view.pageLayout().zoomsToFit():
                view.setZoomFactor(self.zoomFactor)
        if self.position is not None:
            view.setPosition(self.position, False)
        return self

    def save(self, settings):
        """Save the properties that are not None to a QSettings group."""
        if self.pageLayoutMode is not None:
            settings.setValue("pageLayoutMode", self.pageLayoutMode)
        else:
            settings.remove("pageLayoutMode")
        if self.rotation is not None:
            settings.setValue("rotation", self.rotation)
        else:
            settings.remove("rotation")
        if self.orientation is not None:
            settings.setValue("orientation", self.orientation)
        else:
            settings.remove("orientation")
        if self.continuousMode is not None:
            settings.setValue("continuousMode", self.continuousMode)
        else:
            settings.remove("continuousMode")
        if self.viewMode is not None:
            settings.setValue("viewMode", self.viewMode)
        else:
            settings.remove("viewMode")
        if self.zoomFactor is not None:
            settings.setValue("zoomFactor", self.zoomFactor)
        else:
            settings.remove("zoomFactor")
        if self.position is not None:
            settings.setValue("position/pageNumber", self.position.pageNumber)
            settings.setValue("position/x", self.position.x)
            settings.setValue("position/y", self.position.y)
        else:
            settings.remove("position")
        return self

    def load(self, settings):
        """Load the properties from a QSettings group."""
        if settings.contains("pageLayoutMode"):
            v = settings.value("pageLayoutMode", "", str)
            if v:
                self.pageLayoutMode = v
        if settings.contains("rotation"):
            v = settings.value("rotation", -1, int)
            if v in (Rotate_0, Rotate_90, Rotate_180, Rotate_270):
                self.rotation = v
        if settings.contains("orientation"):
            v = settings.value("orientation", 0, int)
            if v in (Horizontal, Vertical):
                self.orientation = v
        if settings.contains("continuousMode"):
            v = settings.value("continuousMode", True, bool)
            self.continuousMode = v
        if settings.contains("viewMode"):
            v = settings.value("viewMode", -1, int)
            if v in (FixedScale, FitHeight, FitWidth, FitBoth):
                self.viewMode = v
        if settings.contains("zoomFactor"):
            v = settings.value("zoomFactor", 0, float)
            if v:
                self.zoomFactor = v
        if settings.contains("position/pageNumber"):
            pageNumber = settings.value("position/pageNumber", -1, int)
            if pageNumber != -1:
                x = settings.value("position/x", 0.0, float)
                y = settings.value("position/y", 0.0, float)
                self.position = Position(pageNumber, x, y)
        return self


class DocumentPropertyStore:
    """Store ViewProperties (settings) on a per-Document basis.

    If you create a DocumentPropertyStore and install it in the
    documentPropertyStore attribute of a View, the View will automatically
    remember its settings for earlier displayed Document instances.

    """

    default = None
    mask = None

    def __init__(self):
        self._properties = weakref.WeakKeyDictionary()

    def get(self, document):
        """Get the View properties stored for the document, if available.

        If a ViewProperties instance is stored in the `default` attribute,
        it is returned when no properties were available. Otherwise, None
        is returned.

        """
        props = self._properties.get(document)
        if props is None:
            if self.default:
                props = self.default
                if self.mask:
                    props = props.copy().mask(self.mask)
        return props

    def set(self, document, properties):
        """Store the View properties for the document.

        If the `mask` attribute is set to a list or tuple of names, only the
        listed properties are remembered.

        """
        if self.mask:
            properties.mask(self.mask)
        self._properties[document] = properties




# -*- coding: utf-8 -*-
#
# This file is part of the qpageview package.
#
# Copyright (c) 2019 - 2019 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.


"""
Generic Link class and handling of links (clickable areas on a Page).

The link area is in coordinates between 0.0 and 1.0, like Poppler does it.
This way we can easily compute where the link area is on a page in different
sizes or rotations.

"""

import collections

from PyQt5.QtCore import pyqtSignal, QEvent, QRectF, Qt

from . import page
from . import rectangles

Area = collections.namedtuple("Area", "left top right bottom")


class Link:
    url = ""
    tooltip = ""
    area = Area(0, 0, 0, 0)

    def __init__(self, left, top, right, bottom, url=None, tooltip=None):
        self.area = Area(left, top, right, bottom)
        if url:
            self.url = url
        if tooltip:
            self.tooltip = tooltip

    def rect(self):
        """Return the area attribute as a QRectF()."""
        r = QRectF()
        r.setCoords(*self.area)
        return r


class Links(rectangles.Rectangles):
    """Manages a list of Link objects.

    See the rectangles documentation for how to access the links.

    """
    def get_coords(self, link):
        return link.area


class LinkViewMixin:
    """Mixin class to enhance view.View with link capabilities."""

    #: (page, link) emitted when the user hovers a link
    linkHovered = pyqtSignal(page.AbstractPage, Link)

    #: (no args) emitted when the user does not hover a link anymore
    linkLeft = pyqtSignal()

    #: (event, page, link) emitted when the user clicks a link
    linkClicked = pyqtSignal(QEvent, page.AbstractPage, Link)

    #: (event, page, link) emitted when a What's This or Toolip is requested.
    #: The event's type determines the type of this help event.
    linkHelpRequested = pyqtSignal(QEvent, page.AbstractPage, Link)

    #: whether to actually enable Link handling
    linksEnabled = True

    def __init__(self, parent=None, **kwds):
        self._currentLinkId = None
        self._linkHighlighter = None
        super().__init__(parent, **kwds)

    def setLinkHighlighter(self, highlighter):
        """Sets a Highlighter (see highlight.py) to highlight a link on hover.

        Use None to remove an active Highlighter. By default no highlighter is
        set to highlight links on hover.

        To be able to actually *use* highlighting, be sure to also mix in the
        HighlightViewMixin class from the highlight module.

        """
        self._linkHighlighter = highlighter

    def linkHighlighter(self):
        """Return the currently set Highlighter, if any.

        By default no highlighter is set to highlight links on hover, and None
        is returned in that case.

        """
        return self._linkHighlighter

    def adjustCursor(self, pos):
        """Adjust the cursor if pos is on a link (and linksEnabled is True).

        Also emits signals when the cursor enters or leaves a link.

        """
        if self.linksEnabled:
            page, link = self.linkAt(pos)
            if link:
                lid = id(link)
            else:
                lid = None
            if lid != self._currentLinkId:
                if self._currentLinkId is not None:
                    self.linkHoverLeave()
                self._currentLinkId = lid
                if lid is not None:
                    self.linkHoverEnter(page, link)
            if link:
                return # do not call super() if we are on a link
        super().adjustCursor(pos)

    def linkAt(self, pos):
        """If the pos (in the viewport) is over a link, return a (page, link) tuple.

        Otherwise returns (None, None).

        """
        pos = pos - self.layoutPosition()
        page = self._pageLayout.pageAt(pos)
        if page:
            links = page.linksAt(pos - page.pos())
            if links:
                return page, links[0]
        return None, None

    def linkHoverEnter(self, page, link):
        """Called when the mouse hovers over a link.

        The default implementation emits the linkHovered(page, link) signal,
        sets a pointing hand mouse cursor, and, if a Highlighter was set using
        setLinkHighlighter(), highlights the link. You can reimplement this
        method to do something different.

        """
        self.setCursor(Qt.PointingHandCursor)
        self.linkHovered.emit(page, link)
        if self._linkHighlighter:
            self.highlight({page: [link.rect()]}, self._linkHighlighter, 3000)

    def linkHoverLeave(self):
        """Called when the mouse does not hover a link anymore.

        The default implementation emits the linkLeft() signal, sets a default
        mouse cursor, and, if a Highlighter was set using setLinkHighlighter(),
        removes the highlighting of the current link. You can reimplement this
        method to do something different.

        """
        self.unsetCursor()
        self.linkLeft.emit()
        if self._linkHighlighter:
            self.clearHighlight(self._linkHighlighter)

    def linkClickEvent(self, ev, page, link):
        """Called when a link is clicked.

        The default implementation emits the linkClicked(event, page, link)
        signal. The event can be used for things like determining which button
        was used, and which keyboard modifiers were in effect.

        """
        self.linkClicked.emit(ev, page, link)

    def linkHelpEvent(self, ev, page, link):
        """Called when a ToolTip or WhatsThis wants to appear.

        The default implementation emits the linkHelpRequested(event, page, link)
        signal. Using the event you can find the position, and the type of the
        help event.

        """
        self.linkHelpRequested.emit(ev, page, link)

    def event(self, ev):
        """Reimplemented to handle HelpEvent for links."""
        if self.linksEnabled and ev.type() in (QEvent.ToolTip, QEvent.WhatsThis):
            page, link = self.linkAt(ev.pos())
            if link:
                self.linkHelpEvent(ev, page, link)
                return True
        return super().event(ev)

    def mousePressEvent(self, ev):
        """Implemented to detect clicking a link and calling linkClickEvent()."""
        if self.linksEnabled:
            page, link = self.linkAt(ev.pos())
            if link:
                self.linkClickEvent(ev, page, link)
                return
        super().mousePressEvent(ev)

    def leaveEvent(self, ev):
        """Implemented to leave a link, might there still be one hovered."""
        if self.linksEnabled and self._currentLinkId is not None:
            self.linkHoverLeave()
            self._currentLinkId = None
        super().leaveEvent(ev)


